#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Endpoint definitions for projects."""

from typing import Union
from uuid import UUID

from fastapi import APIRouter, Depends, Security

from zenml.constants import (
    API,
    PROJECTS,
    STATISTICS,
    VERSION_1,
    WORKSPACES,
)
from zenml.models import (
    Page,
    PipelineFilter,
    PipelineRunFilter,
    ProjectFilter,
    ProjectRequest,
    ProjectResponse,
    ProjectStatistics,
    ProjectUpdate,
)
from zenml.zen_server.auth import AuthContext, authorize
from zenml.zen_server.exceptions import error_response
from zenml.zen_server.feature_gate.endpoint_utils import (
    report_decrement,
)
from zenml.zen_server.rbac.endpoint_utils import (
    verify_permissions_and_create_entity,
    verify_permissions_and_delete_entity,
    verify_permissions_and_get_entity,
    verify_permissions_and_list_entities,
    verify_permissions_and_update_entity,
)
from zenml.zen_server.rbac.models import ResourceType
from zenml.zen_server.rbac.utils import (
    get_allowed_resource_ids,
)
from zenml.zen_server.utils import (
    async_fastapi_endpoint_wrapper,
    make_dependable,
    server_config,
    zen_store,
)

workspace_router = APIRouter(
    prefix=API + VERSION_1 + WORKSPACES,
    tags=["workspaces"],
    responses={401: error_response},
)

router = APIRouter(
    prefix=API + VERSION_1 + PROJECTS,
    tags=["projects"],
    responses={401: error_response},
)


# TODO: kept for backwards compatibility only; to be removed after the migration
@workspace_router.get(
    "",
    responses={401: error_response, 404: error_response, 422: error_response},
    deprecated=True,
)
@router.get(
    "",
    responses={401: error_response, 404: error_response, 422: error_response},
)
@async_fastapi_endpoint_wrapper
def list_projects(
    project_filter_model: ProjectFilter = Depends(
        make_dependable(ProjectFilter)
    ),
    hydrate: bool = False,
    _: AuthContext = Security(authorize),
) -> Page[ProjectResponse]:
    """Lists all projects in the organization.

    Args:
        project_filter_model: Filter model used for pagination, sorting,
            filtering,
        hydrate: Flag deciding whether to hydrate the output model(s)
            by including metadata fields in the response.

    Returns:
        A list of projects.
    """
    return verify_permissions_and_list_entities(
        filter_model=project_filter_model,
        resource_type=ResourceType.PROJECT,
        list_method=zen_store().list_projects,
        hydrate=hydrate,
    )


# TODO: kept for backwards compatibility only; to be removed after the migration
@workspace_router.post(
    "",
    responses={401: error_response, 409: error_response, 422: error_response},
    deprecated=True,
)
@router.post(
    "",
    responses={401: error_response, 409: error_response, 422: error_response},
)
@async_fastapi_endpoint_wrapper
def create_project(
    project_request: ProjectRequest,
    _: AuthContext = Security(authorize),
) -> ProjectResponse:
    """Creates a project based on the requestBody.

    # noqa: DAR401

    Args:
        project_request: Project to create.

    Returns:
        The created project.
    """
    return verify_permissions_and_create_entity(
        request_model=project_request,
        create_method=zen_store().create_project,
    )


# TODO: kept for backwards compatibility only; to be removed after the migration
@workspace_router.get(
    "/{project_name_or_id}",
    responses={401: error_response, 404: error_response, 422: error_response},
    deprecated=True,
)
@router.get(
    "/{project_name_or_id}",
    responses={401: error_response, 404: error_response, 422: error_response},
)
@async_fastapi_endpoint_wrapper
def get_project(
    project_name_or_id: Union[str, UUID],
    hydrate: bool = True,
    _: AuthContext = Security(authorize),
) -> ProjectResponse:
    """Get a project for given name.

    # noqa: DAR401

    Args:
        project_name_or_id: Name or ID of the project.
        hydrate: Flag deciding whether to hydrate the output model(s)
            by including metadata fields in the response.

    Returns:
        The requested project.
    """
    return verify_permissions_and_get_entity(
        id=project_name_or_id,
        get_method=zen_store().get_project,
        hydrate=hydrate,
    )


# TODO: kept for backwards compatibility only; to be removed after the migration
@workspace_router.put(
    "/{project_name_or_id}",
    responses={401: error_response, 404: error_response, 422: error_response},
    deprecated=True,
)
@router.put(
    "/{project_name_or_id}",
    responses={401: error_response, 404: error_response, 422: error_response},
)
@async_fastapi_endpoint_wrapper
def update_project(
    project_name_or_id: UUID,
    project_update: ProjectUpdate,
    _: AuthContext = Security(authorize),
) -> ProjectResponse:
    """Get a project for given name.

    # noqa: DAR401

    Args:
        project_name_or_id: Name or ID of the project to update.
        project_update: the project to use to update

    Returns:
        The updated project.
    """
    return verify_permissions_and_update_entity(
        id=project_name_or_id,
        update_model=project_update,
        get_method=zen_store().get_project,
        update_method=zen_store().update_project,
    )


# TODO: kept for backwards compatibility only; to be removed after the migration
@workspace_router.delete(
    "/{project_name_or_id}",
    responses={401: error_response, 404: error_response, 422: error_response},
    deprecated=True,
)
@router.delete(
    "/{project_name_or_id}",
    responses={401: error_response, 404: error_response, 422: error_response},
)
@async_fastapi_endpoint_wrapper
def delete_project(
    project_name_or_id: Union[str, UUID],
    _: AuthContext = Security(authorize),
) -> None:
    """Deletes a project.

    Args:
        project_name_or_id: Name or ID of the project.
    """
    project = verify_permissions_and_delete_entity(
        id=project_name_or_id,
        get_method=zen_store().get_project,
        delete_method=zen_store().delete_project,
    )
    if server_config().feature_gate_enabled:
        if ResourceType.PROJECT in server_config().reportable_resources:
            report_decrement(ResourceType.PROJECT, resource_id=project.id)


# TODO: kept for backwards compatibility only; to be removed after the migration
@workspace_router.get(
    "/{project_name_or_id}" + STATISTICS,
    responses={401: error_response, 404: error_response, 422: error_response},
    deprecated=True,
)
@router.get(
    "/{project_name_or_id}" + STATISTICS,
    responses={401: error_response, 404: error_response, 422: error_response},
)
@async_fastapi_endpoint_wrapper
def get_project_statistics(
    project_name_or_id: Union[str, UUID],
    auth_context: AuthContext = Security(authorize),
) -> ProjectStatistics:
    """Gets statistics of a project.

    # noqa: DAR401

    Args:
        project_name_or_id: Name or ID of the project to get statistics for.
        auth_context: Authentication context.

    Returns:
        Project statistics.
    """
    project = verify_permissions_and_get_entity(
        id=project_name_or_id,
        get_method=zen_store().get_project,
    )

    user_id = auth_context.user.id

    run_filter = PipelineRunFilter(project=project.id)
    run_filter.configure_rbac(
        authenticated_user_id=user_id,
        id=get_allowed_resource_ids(
            resource_type=ResourceType.PIPELINE_RUN, project_id=project.id
        ),
    )

    pipeline_filter = PipelineFilter(project=project.id)
    pipeline_filter.configure_rbac(
        authenticated_user_id=user_id,
        id=get_allowed_resource_ids(
            resource_type=ResourceType.PIPELINE, project_id=project.id
        ),
    )

    return ProjectStatistics(
        pipelines=zen_store().count_pipelines(filter_model=pipeline_filter),
        runs=zen_store().count_runs(filter_model=run_filter),
    )



#  Copyright (c) ZenML GmbH 2025. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Dynamic pipeline definition."""

from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    List,
    Optional,
    Type,
)

from pydantic import BaseModel, ConfigDict, create_model

from zenml.client import Client
from zenml.config.source import Source
from zenml.enums import ExecutionMode
from zenml.execution.pipeline.utils import (
    should_prevent_pipeline_execution,
)
from zenml.logger import get_logger
from zenml.models import PipelineRunResponse
from zenml.pipelines.pipeline_definition import Pipeline
from zenml.steps.utils import (
    parse_return_type_annotations,
)
from zenml.utils import source_utils

if TYPE_CHECKING:
    from zenml.steps import BaseStep

logger = get_logger(__name__)


class DynamicPipeline(Pipeline):
    """Dynamic pipeline class."""

    def __init__(
        self,
        *,
        depends_on: Optional[List["BaseStep"]] = None,
        **kwargs: Any,
    ) -> None:
        """Initialize the pipeline.

        Args:
            depends_on: The steps that the pipeline depends on.
            **kwargs: Pipeline constructor keyword arguments.
        """
        # This is the only execution mode that is currently supported for
        # dynamic pipelines, so we default to it.
        if kwargs.get("execution_mode", None) is None:
            kwargs["execution_mode"] = ExecutionMode.STOP_ON_FAILURE
        super().__init__(**kwargs)
        self._depends_on = depends_on or []
        self._validate_depends_on(self._depends_on)

    def _validate_depends_on(self, depends_on: List["BaseStep"]) -> None:
        """Validates the steps that the pipeline depends on.

        Args:
            depends_on: The steps that the pipeline depends on.

        Raises:
            RuntimeError: If some of the steps in `depends_on` are duplicated.
        """
        static_ids = set()
        for step in depends_on:
            static_id = step._static_id
            if static_id in static_ids:
                raise RuntimeError(
                    f"The pipeline {self.name} depends on the same step "
                    f"({step.name}) multiple times. To fix this, remove the "
                    "duplicate from the `depends_on` list. You can pass the "
                    "same step function with multiple configurations by using "
                    "the `step.with_options(...)` method."
                )

            static_ids.add(static_id)

    @property
    def depends_on(self) -> List["BaseStep"]:
        """The steps that the pipeline depends on.

        Returns:
            The steps that the pipeline depends on.
        """
        return self._depends_on

    @property
    def is_dynamic(self) -> bool:
        """If the pipeline is dynamic.

        Returns:
            If the pipeline is dynamic.
        """
        return True

    def resolve(self) -> "Source":
        """Resolves the pipeline.

        Raises:
            RuntimeError: If the resolved source is not loadable for dynamic
                pipelines.

        Returns:
            The pipeline source.
        """
        source = super().resolve()
        # We need to validate that the source is loadable for dynamic
        # pipelines as the orchestration environment will need to load the
        # source.
        try:
            source_utils.load(source)
        except Exception as e:
            raise RuntimeError(
                "Unable to resolve dynamic pipeline source. Make sure "
                "your pipeline is defined at the top level of your module."
            ) from e

        return source

    def _prepare_invocations(self, **kwargs: Any) -> None:
        """Prepares the invocations of the pipeline.

        Args:
            **kwargs: Keyword arguments.
        """
        for step in self._depends_on:
            self.add_step_invocation(
                step,
                input_artifacts={},
                external_artifacts={},
                model_artifacts_or_metadata={},
                client_lazy_loaders={},
                parameters={},
                default_parameters={},
                upstream_steps=set(),
            )

    def __call__(
        self, *args: Any, **kwargs: Any
    ) -> Optional[PipelineRunResponse]:
        """Run the pipeline on the active stack.

        Args:
            *args: Entrypoint function arguments.
            **kwargs: Entrypoint function keyword arguments.

        Raises:
            RuntimeError: If the active orchestrator does not support running
                dynamic pipelines.

        Returns:
            The pipeline run or `None` if running with a schedule.
        """
        if should_prevent_pipeline_execution():
            logger.info("Preventing execution of pipeline '%s'.", self.name)
            return None

        stack = Client().active_stack
        if not stack.orchestrator.supports_dynamic_pipelines:
            raise RuntimeError(
                f"The {stack.orchestrator.__class__.__name__} does not "
                "support dynamic pipelines. "
            )

        self.prepare(*args, **kwargs)
        return self._run()

    def _compute_output_schema(self) -> Optional[Dict[str, Any]]:
        """Computes the output schema for the pipeline.

        Returns:
            The output schema for the pipeline.
        """
        try:
            outputs = parse_return_type_annotations(self.entrypoint)
            model_fields: Dict[str, Any] = {
                name: (output.resolved_annotation, ...)
                for name, output in outputs.items()
            }
            output_model: Type[BaseModel] = create_model(
                "PipelineOutput",
                __config__=ConfigDict(extra="forbid"),
                **model_fields,
            )
            return output_model.model_json_schema(mode="serialization")
        except Exception as e:
            logger.debug(
                f"Failed to generate the output schema for pipeline "
                f"`{self.name}: {e}. This means that the pipeline cannot be "
                "deployed.",
            )
            return None



"""Add v2 of trigger schemas [3855b5d051cd].

Revision ID: 3855b5d051cd
Revises: 97109a4a8d26
Create Date: 2026-01-28 13:07:51.002374

"""

import sqlalchemy as sa
import sqlmodel
from alembic import op

# revision identifiers, used by Alembic.
revision = "3855b5d051cd"
down_revision = "97109a4a8d26"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "trigger",
        sa.Column("id", sqlmodel.sql.sqltypes.GUID(), nullable=False),
        sa.Column("created", sa.DateTime(), nullable=False),
        sa.Column("updated", sa.DateTime(), nullable=False),
        sa.Column("name", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("active", sa.Boolean(), nullable=False),
        sa.Column("is_archived", sa.Boolean(), nullable=False),
        sa.Column("type", sa.VARCHAR(length=20), nullable=False),
        sa.Column("flavor", sa.VARCHAR(length=50), nullable=False),
        sa.Column(
            "configuration",
            sa.TEXT(),
            nullable=False,
        ),
        sa.Column("project_id", sqlmodel.sql.sqltypes.GUID(), nullable=False),
        sa.Column("user_id", sqlmodel.sql.sqltypes.GUID(), nullable=True),
        sa.Column("next_occurrence", sa.DateTime(), nullable=True),
        sa.ForeignKeyConstraint(
            ["project_id"],
            ["project.id"],
            name="fk_trigger_project_id_project",
            ondelete="CASCADE",
        ),
        sa.ForeignKeyConstraint(
            ["user_id"],
            ["user.id"],
            name="fk_trigger_user_id_user",
            ondelete="SET NULL",
        ),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint(
            "name", "project_id", name="unique_trigger_name_in_project"
        ),
    )
    with op.batch_alter_table("trigger", schema=None) as batch_op:
        batch_op.create_index(
            "ix_trigger_flavor_next_occurrence",
            ["flavor", "next_occurrence"],
            unique=False,
        )
        batch_op.create_index("ix_trigger_type", ["type"], unique=False)

    op.create_table(
        "trigger_snapshot",
        sa.Column("trigger_id", sqlmodel.sql.sqltypes.GUID(), nullable=False),
        sa.Column("snapshot_id", sqlmodel.sql.sqltypes.GUID(), nullable=False),
        sa.Column("created_at", sa.DateTime(), nullable=False),
        sa.ForeignKeyConstraint(
            ["snapshot_id"],
            ["pipeline_snapshot.id"],
            ondelete="CASCADE",
            name="fk_trigger_snapshot_snapshot_id_pipeline_snapshot",
        ),
        sa.ForeignKeyConstraint(
            ["trigger_id"],
            ["trigger.id"],
            ondelete="CASCADE",
            name="fk_trigger_snapshot_trigger_id_trigger",
        ),
        sa.PrimaryKeyConstraint("trigger_id", "snapshot_id"),
        sa.UniqueConstraint(
            "trigger_id", "snapshot_id", name="unique_trigger_snapshot_link"
        ),
    )
    op.create_table(
        "trigger_execution",
        sa.Column("trigger_id", sqlmodel.sql.sqltypes.GUID(), nullable=False),
        sa.Column(
            "pipeline_run_id", sqlmodel.sql.sqltypes.GUID(), nullable=False
        ),
        sa.Column("created_at", sa.DateTime(), nullable=False),
        sa.ForeignKeyConstraint(
            ["pipeline_run_id"],
            ["pipeline_run.id"],
            ondelete="CASCADE",
            name="fk_trigger_execution_pipeline_run_id_pipeline_run",
        ),
        sa.ForeignKeyConstraint(
            ["trigger_id"],
            ["trigger.id"],
            ondelete="CASCADE",
            name="fk_trigger_execution_trigger_id_trigger",
        ),
        sa.PrimaryKeyConstraint("trigger_id", "pipeline_run_id"),
        sa.UniqueConstraint(
            "trigger_id", "pipeline_run_id", name="unique_trigger_execution"
        ),
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("trigger_execution")
    op.drop_table("trigger_snapshot")
    with op.batch_alter_table("trigger", schema=None) as batch_op:
        batch_op.drop_index("ix_trigger_type")
        batch_op.drop_index("ix_trigger_flavor_next_occurrence")

    op.drop_table("trigger")
    # ### end Alembic commands ###



"""Add is_archived column for schedules [ff6b5e42a915].

Revision ID: ff6b5e42a915
Revises: 514b9fae1500
Create Date: 2025-12-12 11:05:11.954042

"""

import sqlalchemy as sa
from alembic import op

revision = "ff6b5e42a915"
down_revision = "0.93.0"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    with op.batch_alter_table("schedule", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column("is_archived", sa.Boolean(), nullable=True)
        )

    connection = op.get_bind()

    meta = sa.MetaData()
    meta.reflect(bind=connection, only=["schedule"])
    schedule_table = sa.Table("schedule", meta)

    connection.execute(sa.update(schedule_table).values(is_archived=False))

    with op.batch_alter_table("schedule", schema=None) as batch_op:
        batch_op.alter_column(
            "is_archived", existing_type=sa.Boolean, nullable=False
        )


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    with op.batch_alter_table("schedule", schema=None) as batch_op:
        batch_op.drop_column("is_archived")



"""Release [0.91.0].

Revision ID: 0.91.0
Revises: 24552f3be1f2
Create Date: 2025-10-24 09:00:18.541933

"""

# revision identifiers, used by Alembic.
revision = "0.91.0"
down_revision = "24552f3be1f2"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass



#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""The local artifact store is a local implementation of the artifact store.

In ZenML, the inputs and outputs which go through any step is treated as an
artifact and as its name suggests, an `ArtifactStore` is a place where these
artifacts get stored.
"""

import os
from typing import TYPE_CHECKING, ClassVar, Optional, Set, Type, Union

from pydantic import field_validator

from zenml.artifact_stores import (
    BaseArtifactStore,
    BaseArtifactStoreConfig,
    BaseArtifactStoreFlavor,
)
from zenml.config.global_config import GlobalConfiguration
from zenml.exceptions import ArtifactStoreInterfaceError
from zenml.io.local_filesystem import LocalFilesystem
from zenml.utils import io_utils

if TYPE_CHECKING:
    from uuid import UUID

PathType = Union[bytes, str]


class LocalArtifactStoreConfig(BaseArtifactStoreConfig):
    """Config class for the local artifact store.

    Attributes:
        path: The path to the local artifact store.
    """

    SUPPORTED_SCHEMES: ClassVar[Set[str]] = {""}

    path: str = ""

    @field_validator("path")
    @classmethod
    def ensure_path_local(cls, path: str) -> str:
        """Pydantic validator which ensures that the given path is a local path.

        Args:
            path: The path to validate.

        Returns:
            str: The validated (local) path.

        Raises:
            ArtifactStoreInterfaceError: If the given path is not a local path.
        """
        remote_prefixes = ["gs://", "hdfs://", "s3://", "az://", "abfs://"]
        if any(path.startswith(prefix) for prefix in remote_prefixes):
            raise ArtifactStoreInterfaceError(
                f"The path '{path}' you defined for your local artifact store "
                f"starts with a remote prefix."
            )
        return path

    @property
    def is_local(self) -> bool:
        """Checks if this stack component is running locally.

        Returns:
            True if this config is for a local component, False otherwise.
        """
        return True


class LocalArtifactStore(LocalFilesystem, BaseArtifactStore):
    """Artifact Store for local artifacts.

    All methods are inherited from the default `LocalFilesystem`.
    """

    _path: Optional[str] = None

    @staticmethod
    def get_default_local_path(id_: "UUID") -> str:
        """Returns the default local path for a local artifact store.

        Args:
            id_: The id of the local artifact store.

        Returns:
            str: The default local path.
        """
        return os.path.join(
            GlobalConfiguration().local_stores_path,
            str(id_),
        )

    @property
    def path(self) -> str:
        """Returns the path to the local artifact store.

        If the user has not defined a path in the config, this will create a
        sub-folder in the global config directory.

        Returns:
            The path to the local artifact store.
        """
        if self._path:
            return self._path

        if self.config.path:
            self._path = self.config.path
        else:
            self._path = self.get_default_local_path(self.id)
        io_utils.create_dir_recursive_if_not_exists(self._path)
        return self._path

    @property
    def local_path(self) -> Optional[str]:
        """Returns the local path of the artifact store.

        Returns:
            The local path of the artifact store.
        """
        return self.path

    @property
    def custom_cache_key(self) -> Optional[bytes]:
        """Custom cache key.

        The client ID is returned here to invalidate caching when using the same
        local artifact store on multiple client machines.

        Returns:
            Custom cache key.
        """
        return GlobalConfiguration().user_id.bytes


class LocalArtifactStoreFlavor(BaseArtifactStoreFlavor):
    """Class for the `LocalArtifactStoreFlavor`."""

    @property
    def name(self) -> str:
        """Returns the name of the artifact store flavor.

        Returns:
            str: The name of the artifact store flavor.
        """
        return "local"

    @property
    def docs_url(self) -> Optional[str]:
        """A url to point at docs explaining this flavor.

        Returns:
            A flavor docs url.
        """
        return self.generate_default_docs_url()

    @property
    def sdk_docs_url(self) -> Optional[str]:
        """A url to point at docs explaining this flavor.

        Returns:
            A flavor docs url.
        """
        return self.generate_default_sdk_docs_url()

    @property
    def logo_url(self) -> str:
        """A url to represent the flavor in the dashboard.

        Returns:
            The flavor logo.
        """
        return "https://public-flavor-logos.s3.eu-central-1.amazonaws.com/artifact_store/local.svg"

    @property
    def config_class(self) -> Type[LocalArtifactStoreConfig]:
        """Config class for this flavor.

        Returns:
            The config class.
        """
        return LocalArtifactStoreConfig

    @property
    def implementation_class(self) -> Type[LocalArtifactStore]:
        """Implementation class.

        Returns:
            The implementation class.
        """
        return LocalArtifactStore



#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Alembic environment script.

This script is used by all Alembic CLI commands. It uses the SqlZenStore
currently configured in the global configuration, or the default store if
no store is configured, to get the SQLAlchemy engine and the context to run
migrations.
"""

from typing import cast

from alembic import context
from alembic.runtime.environment import EnvironmentContext

from zenml.config.global_config import GlobalConfiguration
from zenml.enums import StoreType
from zenml.zen_stores.migrations.alembic import Alembic
from zenml.zen_stores.sql_zen_store import SqlZenStore


def run_migrations() -> None:
    """Run migrations through the ZenML SqlZenStore and Alembic classes."""
    store_cfg = GlobalConfiguration().store_configuration

    if store_cfg.type != StoreType.SQL:
        store_cfg = GlobalConfiguration().get_default_store()

    store = SqlZenStore(
        config=store_cfg,
        skip_default_registrations=True,
        skip_migrations=True,
    )

    a = Alembic(engine=store.engine, context=cast(EnvironmentContext, context))
    a.run_migrations(fn=None)


run_migrations()



#  Copyright (c) ZenML GmbH 2024. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Rate limiting for the ZenML Server."""

import inspect
import time
from collections import defaultdict
from contextlib import contextmanager
from functools import wraps
from typing import (
    Any,
    Callable,
    Dict,
    Generator,
    List,
    Optional,
    TypeVar,
    cast,
)

from starlette.requests import Request

from zenml.logger import get_logger
from zenml.zen_server.utils import server_config

logger = get_logger(__name__)
F = TypeVar("F", bound=Callable[..., Any])


class RequestLimiter:
    """Simple in-memory rate limiter."""

    def __init__(
        self,
        day_limit: Optional[int] = None,
        minute_limit: Optional[int] = None,
    ):
        """Initializes the limiter.

        Args:
            day_limit: The number of requests allowed per day.
            minute_limit: The number of requests allowed per minute.

        Raises:
            ValueError: If both day_limit and minute_limit are None.
        """
        self.limiting_enabled = server_config().rate_limit_enabled
        if not self.limiting_enabled:
            return
        if day_limit is None and minute_limit is None:
            raise ValueError("Pass either day or minuter limits, or both.")
        self.day_limit = day_limit
        self.minute_limit = minute_limit
        self.limiter: Dict[str, List[float]] = defaultdict(list)

    def hit_limiter(self, request: Request) -> None:
        """Increase the number of hits in the limiter.

        Args:
            request: Request object.

        Raises:
            HTTPException: If the request limit is exceeded.
        """
        if not self.limiting_enabled:
            return
        from fastapi import HTTPException

        requester = self._get_ipaddr(request)
        now = time.time()
        minute_ago = now - 60
        day_ago = now - 60 * 60 * 24
        self.limiter[requester].append(now)

        from bisect import bisect_left

        # remove failures older than a day
        older_index = bisect_left(self.limiter[requester], day_ago)
        self.limiter[requester] = self.limiter[requester][older_index:]

        if self.day_limit and len(self.limiter[requester]) > self.day_limit:
            raise HTTPException(
                status_code=429, detail="Daily request limit exceeded."
            )
        minute_requests = len(
            [
                limiter_hit
                for limiter_hit in self.limiter[requester][::-1]
                if limiter_hit >= minute_ago
            ]
        )
        if self.minute_limit and minute_requests > self.minute_limit:
            raise HTTPException(
                status_code=429, detail="Minute request limit exceeded."
            )

    def reset_limiter(self, request: Request) -> None:
        """Resets the limiter on successful request.

        Args:
            request: Request object.
        """
        if self.limiting_enabled:
            requester = self._get_ipaddr(request)
            if requester in self.limiter:
                del self.limiter[requester]

    def _get_ipaddr(self, request: Request) -> str:
        """Returns the IP address for the current request.

        Based on the X-Forwarded-For headers or client information.

        Args:
            request: The request object.

        Returns:
            The ip address for the current request (or 127.0.0.1 if none found).
        """
        if "X_FORWARDED_FOR" in request.headers:
            return request.headers["X_FORWARDED_FOR"]
        else:
            if not request.client or not request.client.host:
                return "127.0.0.1"

            return request.client.host

    @contextmanager
    def limit_failed_requests(
        self, request: Request
    ) -> Generator[None, Any, Any]:
        """Limits the number of failed requests.

        Args:
            request: Request object.

        Yields:
            None
        """
        self.hit_limiter(request)

        yield

        # if request was successful - reset limiter
        self.reset_limiter(request)


def rate_limit_requests(
    day_limit: Optional[int] = None,
    minute_limit: Optional[int] = None,
) -> Callable[..., Any]:
    """Decorator to handle exceptions in the API.

    Args:
        day_limit: Number of requests allowed per day.
        minute_limit: Number of requests allowed per minute.

    Returns:
        Decorated function.
    """
    limiter = RequestLimiter(day_limit=day_limit, minute_limit=minute_limit)

    def decorator(func: F) -> F:
        request_arg, request_kwarg = None, None
        parameters = inspect.signature(func).parameters
        for arg_num, arg_name in enumerate(parameters):
            if parameters[arg_name].annotation == Request:
                request_arg = arg_num
                request_kwarg = arg_name
                break
        if request_arg is None or request_kwarg is None:
            raise ValueError(
                "Rate limiting APIs must have argument of `Request` type."
            )

        @wraps(func)
        def decorated(
            *args: Any,
            **kwargs: Any,
        ) -> Any:
            if request_kwarg in kwargs:
                request = kwargs[request_kwarg]
            else:
                request = args[request_arg]
            with limiter.limit_failed_requests(request):
                return func(*args, **kwargs)

        return cast(F, decorated)

    return decorator



#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Kubernetes integration flavors."""

from zenml.integrations.kubernetes.flavors.kubernetes_deployer_flavor import (
    KubernetesDeployerConfig,
    KubernetesDeployerFlavor,
    KubernetesDeployerSettings,
)
from zenml.integrations.kubernetes.flavors.kubernetes_orchestrator_flavor import (
    KubernetesOrchestratorConfig,
    KubernetesOrchestratorFlavor,
    KubernetesOrchestratorSettings,
)
from zenml.integrations.kubernetes.flavors.kubernetes_step_operator_flavor import (
    KubernetesStepOperatorConfig,
    KubernetesStepOperatorFlavor,
    KubernetesStepOperatorSettings,
)

__all__ = [
    "KubernetesDeployerConfig",
    "KubernetesDeployerFlavor",
    "KubernetesDeployerSettings",
    "KubernetesOrchestratorFlavor",
    "KubernetesOrchestratorConfig",
    "KubernetesOrchestratorSettings",
    "KubernetesStepOperatorConfig",
    "KubernetesStepOperatorFlavor",
    "KubernetesStepOperatorSettings",
]



#  Copyright (c) ZenML GmbH 2025. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Utility functions to run a pipeline from the server."""

import hashlib
import os
import sys
import threading
from concurrent.futures import Future, ThreadPoolExecutor
from typing import Any, Callable, Dict, List, Optional, Tuple
from uuid import UUID

from packaging import version

from zenml import LogsRequest
from zenml.analytics.enums import AnalyticsEvent
from zenml.analytics.utils import track_handler
from zenml.config.base_settings import BaseSettings
from zenml.config.pipeline_configurations import PipelineConfiguration
from zenml.config.pipeline_run_configuration import (
    PipelineRunConfiguration,
)
from zenml.config.step_configurations import Step, StepConfigurationUpdate
from zenml.constants import (
    ENV_ZENML_ACTIVE_PROJECT_ID,
    ENV_ZENML_ACTIVE_STACK_ID,
    ENV_ZENML_RUNNER_IMAGE_DISABLE_UV,
    ENV_ZENML_RUNNER_PARENT_IMAGE,
    ENV_ZENML_RUNNER_POD_TIMEOUT,
    LOGS_RUNNER_SOURCE,
    RUN_TEMPLATE_TRIGGERS_FEATURE_NAME,
    handle_bool_env_var,
    handle_int_env_var,
)
from zenml.enums import ExecutionStatus, StackComponentType, StoreType
from zenml.exceptions import IllegalOperationError, MaxConcurrentTasksError
from zenml.logger import get_logger
from zenml.models import (
    CodeReferenceRequest,
    FlavorFilter,
    PipelineBuildResponse,
    PipelineRunResponse,
    PipelineRunTriggerInfo,
    PipelineRunUpdate,
    PipelineSnapshotRequest,
    PipelineSnapshotResponse,
    PipelineSnapshotRunRequest,
    StackResponse,
)
from zenml.pipelines.build_utils import compute_stack_checksum
from zenml.pipelines.run_utils import (
    create_placeholder_run,
    validate_run_config_is_runnable_from_server,
    validate_stack_is_runnable_from_server,
)
from zenml.stack.flavor import Flavor
from zenml.utils import pydantic_utils, requirements_utils, settings_utils
from zenml.zen_server.auth import AuthContext, generate_access_token
from zenml.zen_server.feature_gate.endpoint_utils import (
    report_usage,
)
from zenml.zen_server.pipeline_execution.runner_entrypoint_configuration import (
    RunnerEntrypointConfiguration,
)
from zenml.zen_server.pipeline_execution.workload_manager_interface import (
    WorkloadType,
)
from zenml.zen_server.utils import (
    get_auth_context,
    server_config,
    set_auth_context,
    snapshot_executor,
    workload_manager,
    zen_store,
)

logger = get_logger(__name__)

RUNNER_IMAGE_REPOSITORY = "zenml-runner"


class BoundedThreadPoolExecutor:
    """Thread pool executor which only allows a maximum number of concurrent tasks."""

    def __init__(self, max_workers: int, **kwargs: Any) -> None:
        """Initialize the executor.

        Args:
            max_workers: The maximum number of workers.
            **kwargs: Arguments to pass to the thread pool executor.
        """
        self._executor = ThreadPoolExecutor(max_workers=max_workers, **kwargs)
        self._semaphore = threading.BoundedSemaphore(value=max_workers)

    def submit(
        self, fn: Callable[..., Any], *args: Any, **kwargs: Any
    ) -> Future[Any]:
        """Submit a task to the executor.

        Args:
            fn: The function to execute.
            *args: The arguments to pass to the function.
            **kwargs: The keyword arguments to pass to the function.

        Raises:
            Exception: If the task submission fails.
            MaxConcurrentTasksError: If the maximum number of concurrent tasks
                is reached.

        Returns:
            The future of the task.
        """
        if not self._semaphore.acquire(blocking=False):
            raise MaxConcurrentTasksError(
                "Maximum number of concurrent tasks reached."
            )

        try:
            future = self._executor.submit(fn, *args, **kwargs)
        except Exception:
            self._semaphore.release()
            raise
        else:
            future.add_done_callback(lambda _: self._semaphore.release())
            return future

    def shutdown(self, **kwargs: Any) -> None:
        """Shutdown the executor.

        Args:
            **kwargs: Keyword arguments to pass to the shutdown method of the
                executor.
        """
        self._executor.shutdown(**kwargs)


def run_snapshot(
    snapshot: PipelineSnapshotResponse,
    auth_context: AuthContext,
    request: PipelineSnapshotRunRequest,
    sync: bool = False,
    template_id: Optional[UUID] = None,
    create_new_snapshot: bool = True,
    implicit_auth_context: bool = True,
    wait_runner_pod: bool = True,
    trigger_id: UUID | None = None,
) -> PipelineRunResponse:
    """Run a snapshot from the server.

    Args:
        snapshot: The snapshot to run.
        auth_context: Authentication context.
        request: The run request.
        sync: Whether to run the snapshot synchronously.
        template_id: The ID of the template from which to create the snapshot
            request.
        create_new_snapshot: Whether to create a new, copy snapshot.
        implicit_auth_context: Whether to use implicit auth context or create an explicit new one.
        wait_runner_pod: Whether to wait for runner pod completion.
        trigger_id: The trigger ID that generated the snapshot run (optional).

    Raises:
        MaxConcurrentTasksError: If the maximum number of concurrent run
            snapshot tasks is reached.

    Returns:
        ID of the new pipeline run.
    """
    if not implicit_auth_context:
        set_auth_context(auth_context)
    logger.info("Current auth context: %s", get_auth_context())
    build, stack, zenml_version = validate_snapshot_for_server_execution(
        snapshot=snapshot,
        run_configuration=request.run_configuration,
    )

    snapshot_request = snapshot_request_from_source_snapshot(
        source_snapshot=snapshot,
        config=request.run_configuration or PipelineRunConfiguration(),
        template_id=template_id,
    )

    ensure_async_orchestrator(snapshot=snapshot_request, stack=stack)

    if create_new_snapshot:
        target_snapshot = zen_store().create_snapshot(snapshot_request)
    else:
        target_snapshot = snapshot

    trigger_info = None
    if request.step_run:
        trigger_info = PipelineRunTriggerInfo(
            step_run_id=request.step_run,
        )

    placeholder_run = create_placeholder_run(
        snapshot=target_snapshot,
        trigger_info=trigger_info,
        logs=LogsRequest(source=LOGS_RUNNER_SOURCE),
    )

    if trigger_id:
        zen_store().create_trigger_execution(
            trigger_id=trigger_id,
            pipeline_run_id=placeholder_run.id,
        )

    report_usage(
        feature=RUN_TEMPLATE_TRIGGERS_FEATURE_NAME,
        resource_id=placeholder_run.id,
    )

    environment = build_runner_environment(
        snapshot=target_snapshot,
        stack=stack,
        run_id=placeholder_run.id,
        auth_context=auth_context,
        zenml_version=zenml_version,
    )
    command = RunnerEntrypointConfiguration.get_entrypoint_command()
    args = RunnerEntrypointConfiguration.get_entrypoint_arguments(
        snapshot_id=target_snapshot.id,
        run_id=placeholder_run.id,
    )
    dockerfile = build_runner_dockerfile(
        stack=stack, build=build, zenml_version=zenml_version
    )

    workload_id = placeholder_run.id if trigger_id else target_snapshot.id
    workload_type = WorkloadType.RUN if trigger_id else WorkloadType.SNAPSHOT

    def _task() -> None:
        with track_handler(
            event=AnalyticsEvent.RUN_PIPELINE
        ) as analytics_handler:
            analytics_handler.metadata = get_pipeline_run_analytics_metadata(
                snapshot=target_snapshot,
                stack=stack,
                source_snapshot_id=snapshot.id,
                run_id=placeholder_run.id,
            )

            analytics_handler.metadata["trigger_execution"] = (
                True if trigger_id else False
            )

            try:
                _build_and_run(
                    workload_id=workload_id,
                    workload_type=workload_type,
                    command=command,
                    arguments=args,
                    environment=environment,
                    dockerfile=dockerfile,
                    wait_for_completion=wait_runner_pod,
                    success_message="Pipeline run started successfully.",
                )
            except Exception:
                logger.exception(
                    "Failed to run snapshot %s, run ID: %s",
                    str(snapshot.id),
                    str(placeholder_run.id),
                )
                run_status = zen_store().get_run_status(placeholder_run.id)
                if run_status == ExecutionStatus.INITIALIZING:
                    # The run isn't in the provisioning status yet, which means
                    # the orchestrator wasn't able to submit the run. In this
                    # case, we can update the run to failed.
                    zen_store().update_run(
                        run_id=placeholder_run.id,
                        run_update=PipelineRunUpdate(
                            status=ExecutionStatus.FAILED,
                            status_reason="Failed to start run.",
                        ),
                    )
                    raise

    if sync:
        _task()
    else:
        try:
            snapshot_executor().submit(_task)
        except MaxConcurrentTasksError:
            zen_store().delete_run(run_id=placeholder_run.id)
            raise MaxConcurrentTasksError(
                "Maximum number of concurrent snapshot tasks reached."
            ) from None

    return placeholder_run


def resume_run(run: PipelineRunResponse) -> Future[None]:
    """Resume a run from the server.

    Args:
        run: The pipeline run that should be resumed

    Raises:
        MaxConcurrentTasksError: If workload submission exceeds concurrency
            limits.
        IllegalOperationError: If the run is not resuming.

    Returns:
        A future that resolves when the run is resumed.
    """
    if run.status != ExecutionStatus.RESUMING:
        raise IllegalOperationError(
            "Cannot restart a run that is not resuming."
        )

    if not server_config().workload_manager_enabled:
        raise IllegalOperationError(
            "Resuming runs is only possible when the workload manager is enabled."
        )

    snapshot = run.snapshot
    if not snapshot or not snapshot.runnable:
        raise IllegalOperationError(
            "Cannot resume a run that is not based on a runnable snapshot."
        )
    if not snapshot.is_dynamic:
        raise IllegalOperationError(
            "Cannot resume a run of a static pipeline."
        )

    has_runner_logs = any(
        log.source == LOGS_RUNNER_SOURCE for log in run.log_collection or []
    )
    if not has_runner_logs:
        # TODO: When the same run gets resumed multiple times, or a snapshot
        # run gets resumed, the logs get appended to the file.
        zen_store().update_run(
            run.id,
            PipelineRunUpdate(
                add_logs=[LogsRequest(source=LOGS_RUNNER_SOURCE)]
            ),
        )

    auth_context = get_auth_context()
    assert auth_context

    build, stack, zenml_version = validate_snapshot_for_server_execution(
        snapshot=snapshot
    )
    environment = build_runner_environment(
        snapshot=snapshot,
        stack=stack,
        run_id=run.id,
        auth_context=auth_context,
        zenml_version=zenml_version,
    )
    command = RunnerEntrypointConfiguration.get_entrypoint_command()
    args = RunnerEntrypointConfiguration.get_entrypoint_arguments(
        snapshot_id=snapshot.id,
        run_id=run.id,
        resume=True,
    )
    dockerfile = build_runner_dockerfile(
        stack=stack, build=build, zenml_version=zenml_version
    )

    def _task() -> None:
        _build_and_run(
            workload_id=run.id,
            workload_type=WorkloadType.RUN,
            command=command,
            arguments=args,
            environment=environment,
            dockerfile=dockerfile,
            wait_for_completion=True,
        )

    try:
        future = snapshot_executor().submit(_task)
    except MaxConcurrentTasksError:
        raise MaxConcurrentTasksError(
            "Maximum number of concurrent snapshot tasks reached."
        ) from None

    return future


def ensure_async_orchestrator(
    snapshot: PipelineSnapshotRequest, stack: StackResponse
) -> None:
    """Ensures the orchestrator is configured to run async.

    Args:
        snapshot: Snapshot request in which the orchestrator
            configuration should be updated to ensure the orchestrator is
            running async.
        stack: The stack on which the snapshot will run.
    """
    orchestrator = stack.components[StackComponentType.ORCHESTRATOR][0]
    flavors = zen_store().list_flavors(
        FlavorFilter(name=orchestrator.flavor_name, type=orchestrator.type)
    )
    flavor = Flavor.from_model(flavors[0])

    if "synchronous" in flavor.config_class.model_fields:
        key = settings_utils.get_flavor_setting_key(flavor)

        if settings := snapshot.pipeline_configuration.settings.get(key):
            settings_dict = settings.model_dump()
        else:
            settings_dict = {}

        settings_dict["synchronous"] = False
        snapshot.pipeline_configuration.settings[key] = (
            BaseSettings.model_validate(settings_dict)
        )


def generate_image_hash(dockerfile: str) -> str:
    """Generate a hash of the Dockerfile.

    Args:
        dockerfile: The Dockerfile for which to generate the hash.

    Returns:
        The hash of the Dockerfile.
    """
    hash_ = hashlib.md5()  # nosec
    # Uncomment this line when developing to guarantee a new docker image gets
    # built after restarting the server
    # hash_.update(f"{os.getpid()}".encode())
    hash_.update(dockerfile.encode())
    return hash_.hexdigest()


def generate_dockerfile(
    pypi_requirements: List[str],
    apt_packages: List[str],
    zenml_version: str,
    python_version: str,
) -> str:
    """Generate a Dockerfile that installs the requirements.

    Args:
        pypi_requirements: The PyPI requirements to install.
        apt_packages: The APT packages to install.
        zenml_version: The ZenML version to use as parent image.
        python_version: The Python version to use as parent image.

    Returns:
        The Dockerfile.
    """
    parent_image = os.environ.get(
        ENV_ZENML_RUNNER_PARENT_IMAGE,
        f"zenmldocker/zenml:{zenml_version}-py{python_version}",
    )

    lines = [f"FROM {parent_image}"]
    if apt_packages:
        apt_packages_string = " ".join(f"'{p}'" for p in apt_packages)
        lines.append(
            "RUN apt-get update && apt-get install -y "
            f"--no-install-recommends {apt_packages_string}"
        )

    if pypi_requirements:
        pypi_requirements_string = " ".join(
            [f"'{r}'" for r in pypi_requirements]
        )

        if handle_bool_env_var(
            ENV_ZENML_RUNNER_IMAGE_DISABLE_UV, default=False
        ):
            lines.append(
                f"RUN pip install --default-timeout=60 --no-cache-dir "
                f"{pypi_requirements_string}"
            )
        else:
            lines.append("RUN pip install uv")
            lines.append(
                f"RUN uv pip install --no-cache-dir {pypi_requirements_string}"
            )

    return "\n".join(lines)


def snapshot_request_from_source_snapshot(
    source_snapshot: PipelineSnapshotResponse,
    config: PipelineRunConfiguration,
    template_id: Optional[UUID] = None,
) -> "PipelineSnapshotRequest":
    """Generate a snapshot request from a source snapshot.

    Args:
        source_snapshot: The source snapshot from which to create the
            snapshot request.
        config: The run configuration.
        template_id: The ID of the template from which to create the snapshot
            request.

    Raises:
        ValueError: If there are missing/extra step parameters in the run
            configuration.

    Returns:
        The generated snapshot request.
    """
    assert source_snapshot.stack
    assert source_snapshot.build

    pipeline_update_exclude = {"name"}
    if not source_snapshot.is_dynamic:
        pipeline_update_exclude.add("parameters")

    if config.settings:
        convert_component_shortcut_settings_keys(
            settings=config.settings, stack=source_snapshot.stack
        )

    pipeline_update = config.model_dump(
        include=set(PipelineConfiguration.model_fields),
        exclude=pipeline_update_exclude,
        exclude_unset=True,
        exclude_none=True,
    )
    if pipeline_secrets := pipeline_update.get("secrets", []):
        pipeline_update["secrets"] = [
            zen_store().get_secret_by_name_or_id(secret).id
            for secret in pipeline_secrets
        ]
    pipeline_configuration = pydantic_utils.update_model(
        source_snapshot.pipeline_configuration, pipeline_update
    )

    pipeline_spec = source_snapshot.pipeline_spec
    if pipeline_spec and pipeline_configuration.parameters:
        # Also include the updated pipeline parameters in the pipeline spec, as
        # the frontend and some other code still relies on the parameters in it
        pipeline_spec = pipeline_spec.model_copy(
            update={"parameters": pipeline_configuration.parameters}
        )

    steps = {}
    step_config_updates = config.steps or {}
    for invocation_id, step in source_snapshot.step_configurations.items():
        step_update_model = step_config_updates.get(
            invocation_id, StepConfigurationUpdate()
        )
        if step_update_model.settings:
            convert_component_shortcut_settings_keys(
                settings=step_update_model.settings,
                stack=source_snapshot.stack,
            )
        step_update = step_update_model.model_dump(
            # Get rid of deprecated name to prevent overriding the step name
            # with `None`.
            exclude={"name"},
            exclude_unset=True,
            exclude_none=True,
        )
        if step_secrets := step_update.get("secrets", []):
            step_update["secrets"] = [
                zen_store().get_secret_by_name_or_id(secret).id
                for secret in step_secrets
            ]
        step_config = pydantic_utils.update_model(
            step.step_config_overrides, step_update
        )
        merged_step_config = step_config.apply_pipeline_configuration(
            pipeline_configuration
        )

        required_parameters = set(step.config.parameters)
        configured_parameters = set(step_config.parameters)

        unknown_parameters = configured_parameters - required_parameters
        if unknown_parameters:
            raise ValueError(
                "Run configuration contains the following unknown "
                f"parameters for step {invocation_id}: {unknown_parameters}."
            )

        missing_parameters = required_parameters - configured_parameters
        if missing_parameters:
            raise ValueError(
                "Run configuration is missing the following required "
                f"parameters for step {invocation_id}: {missing_parameters}."
            )

        steps[invocation_id] = Step(
            spec=step.spec,
            config=merged_step_config,
            step_config_overrides=step_config,
        )

    code_reference_request = None
    if source_snapshot.code_reference:
        code_reference_request = CodeReferenceRequest(
            commit=source_snapshot.code_reference.commit,
            subdirectory=source_snapshot.code_reference.subdirectory,
            code_repository=source_snapshot.code_reference.code_repository.id,
        )

    zenml_version = zen_store().get_store_info().version

    # Compute the source snapshot ID:
    # - If the source snapshot has a name, we use it as the source snapshot.
    #   That way, all runs will be associated with this snapshot.
    # - If the source snapshot is based on another snapshot (which therefore
    #   has a name), we use that one instead.
    # - If the source snapshot does not have a name and is not based on another
    #   snapshot, we don't set a source snapshot.
    #
    # With this, we ensure that all runs are associated with the closest named
    # source snapshot.
    source_snapshot_id = None
    if source_snapshot.name:
        source_snapshot_id = source_snapshot.id
    elif source_snapshot.source_snapshot_id:
        source_snapshot_id = source_snapshot.source_snapshot_id

    return PipelineSnapshotRequest(
        project=source_snapshot.project_id,
        is_dynamic=source_snapshot.is_dynamic,
        run_name_template=config.run_name or source_snapshot.run_name_template,
        pipeline_configuration=pipeline_configuration,
        step_configurations=steps,
        client_environment={},
        client_version=zenml_version,
        server_version=zenml_version,
        stack=source_snapshot.stack.id,
        pipeline=source_snapshot.pipeline.id
        if source_snapshot.pipeline
        else None,
        build=source_snapshot.build.id,
        schedule=None,
        code_reference=code_reference_request,
        code_path=source_snapshot.code_path,
        template=template_id,
        source_snapshot=source_snapshot_id,
        pipeline_version_hash=source_snapshot.pipeline_version_hash,
        pipeline_spec=pipeline_spec,
        source_code=source_snapshot.source_code,
    )


def get_pipeline_run_analytics_metadata(
    snapshot: "PipelineSnapshotResponse",
    stack: StackResponse,
    source_snapshot_id: UUID,
    run_id: UUID,
) -> Dict[str, Any]:
    """Get metadata for the pipeline run analytics event.

    Args:
        snapshot: The snapshot of the run.
        stack: The stack on which the run will happen.
        source_snapshot_id: ID of the source snapshot from which the run
            was started.
        run_id: ID of the run.

    Returns:
        The analytics metadata.
    """
    custom_materializer = False
    for step in snapshot.step_configurations.values():
        for output in step.config.outputs.values():
            for source in output.materializer_source:
                if not source.is_internal:
                    custom_materializer = True

    assert snapshot.user_id
    stack_creator = stack.user_id
    own_stack = stack_creator and stack_creator == snapshot.user_id

    stack_metadata = {
        component_type.value: component_list[0].flavor_name
        for component_type, component_list in stack.components.items()
    }

    return {
        "project_id": snapshot.project_id,
        "store_type": "rest",  # This method is called from within a REST endpoint
        **stack_metadata,
        "total_steps": len(snapshot.step_configurations),
        "schedule": snapshot.schedule is not None,
        "custom_materializer": custom_materializer,
        "own_stack": own_stack,
        "pipeline_run_id": str(run_id),
        "source_snapshot_id": str(source_snapshot_id),
    }


def convert_component_shortcut_settings_keys(
    settings: Dict[str, "BaseSettings"], stack: "StackResponse"
) -> None:
    """Convert component shortcut settings keys.

    Args:
        settings: Dictionary of settings.
        stack: The stack response.

    Raises:
        ValueError: If the shortcut key is ambiguous because the stack has
            multiple components of the same type.
        ValueError: If stack component settings were defined both using the
            full and the shortcut key.
    """
    for component_type, component_list in stack.components.items():
        shortcut_key = str(component_type)
        if component_settings := settings.pop(shortcut_key, None):
            if len(component_list) > 1:
                raise ValueError(
                    "Unable to convert shortcut settings key for stack with "
                    f"multiple components of type {component_type}."
                )

            key = f"{component_type}.{component_list[0].flavor_name}"
            if key in settings:
                raise ValueError(
                    f"Duplicate settings provided for your {shortcut_key} "
                    f"using the keys {shortcut_key} and {key}. Remove settings "
                    "for one of them to fix this error."
                )

            settings[key] = component_settings


def validate_snapshot_for_server_execution(
    snapshot: PipelineSnapshotResponse,
    run_configuration: Optional[PipelineRunConfiguration] = None,
) -> Tuple[PipelineBuildResponse, StackResponse, str]:
    """Validate that a snapshot can be executed from the server.

    Args:
        snapshot: Snapshot to validate.
        run_configuration: Optional run configuration to validate as well.

    Raises:
        ValueError: If the snapshot or run configuration is not runnable from
            the server.

    Returns:
        The snapshot build, stack, and ZenML version.
    """
    if not snapshot.runnable:
        if stack := snapshot.stack:
            validate_stack_is_runnable_from_server(
                zen_store=zen_store(), stack=stack
            )
        if not snapshot.build:
            raise ValueError(
                "This snapshot can not be run via the server because it does "
                "not have an associated build. This is probably because the "
                "build has been deleted."
            )

        raise ValueError("This snapshot can not be run via the server.")

    build = snapshot.build
    assert build
    stack = build.stack
    assert stack

    if build.stack_checksum and build.stack_checksum != compute_stack_checksum(
        stack=stack
    ):
        raise ValueError(
            f"The stack {stack.name} has been updated since it was used for "
            "the snapshot. This means the Docker "
            "images associated with this template most likely do not contain "
            "the necessary requirements. Please create a new snapshot with "
            "the updated stack."
        )

    validate_stack_is_runnable_from_server(zen_store=zen_store(), stack=stack)
    if run_configuration:
        validate_run_config_is_runnable_from_server(
            run_configuration, is_dynamic=snapshot.is_dynamic
        )

    assert build.zenml_version
    return build, stack, build.zenml_version


def build_runner_environment(
    snapshot: PipelineSnapshotResponse,
    stack: StackResponse,
    run_id: UUID,
    auth_context: AuthContext,
    zenml_version: str,
) -> Dict[str, str]:
    """Build environment variables for a runner workload.

    Args:
        snapshot: Snapshot to execute.
        stack: Stack to use for the run.
        run_id: Pipeline run ID scoped into the API token.
        auth_context: Authentication context.
        zenml_version: ZenML version to use.

    Raises:
        RuntimeError: If the server URL is not configured.

    Returns:
        Runner workload environment variables.
    """
    server_url = server_config().server_url
    if not server_url:
        raise RuntimeError(
            "The server URL is not set in the server configuration."
        )

    # We create an API token scoped to the pipeline run that never expires.
    api_token = generate_access_token(
        user_id=auth_context.user.id,
        pipeline_run_id=run_id,
        # Keep the original API key or device scopes, if any.
        api_key=auth_context.api_key,
        device=auth_context.device,
        # Never expire the token.
        expires_in=0,
    ).access_token

    return {
        ENV_ZENML_ACTIVE_PROJECT_ID: str(snapshot.project_id),
        ENV_ZENML_ACTIVE_STACK_ID: str(stack.id),
        "ZENML_VERSION": zenml_version,
        "ZENML_STORE_URL": server_url,
        "ZENML_STORE_TYPE": StoreType.REST.value,
        "ZENML_STORE_API_TOKEN": api_token,
        "ZENML_STORE_VERIFY_SSL": "True",
    }


def build_runner_dockerfile(
    stack: StackResponse,
    build: PipelineBuildResponse,
    zenml_version: str,
) -> str:
    """Build the runner Dockerfile.

    Args:
        stack: Stack to use.
        build: Snapshot build metadata.
        zenml_version: ZenML version to use.

    Returns:
        Dockerfile content for the runner workload.
    """
    if build.python_version:
        version_info = version.parse(build.python_version)
        python_version = f"{version_info.major}.{version_info.minor}"
    else:
        python_version = f"{sys.version_info.major}.{sys.version_info.minor}"

    pypi_requirements, apt_packages = (
        requirements_utils.get_requirements_for_stack(
            stack=stack, python_version=python_version
        )
    )

    return generate_dockerfile(
        pypi_requirements=pypi_requirements,
        apt_packages=apt_packages,
        zenml_version=zenml_version,
        python_version=python_version,
    )


def _build_and_run(
    workload_id: UUID,
    workload_type: WorkloadType,
    command: List[str],
    arguments: List[str],
    environment: Dict[str, str],
    dockerfile: str,
    wait_for_completion: bool,
    success_message: str | None = None,
) -> None:
    """Build and run a runner workload.

    Args:
        workload_id: Workload ID.
        workload_type: Workload type.
        command: Entrypoint command.
        arguments: Entrypoint arguments.
        environment: Workload environment variables.
        dockerfile: Dockerfile for the runner image.
        wait_for_completion: Whether to wait for the runner workload.
        success_message: Optional message logged after a successful start.
    """
    image_hash = generate_image_hash(dockerfile=dockerfile)
    logger.info(
        "Building runner image %s for dockerfile:\n%s", image_hash, dockerfile
    )

    # Building a Docker image with requirements and apt packages from the
    # stack only (no code). Ideally, only orchestrator requirements should
    # be added to the Docker image, but we have to instantiate the entire
    # stack to get the orchestrator to run pipelines.
    runner_image = workload_manager().build_and_push_image(
        workload_id=workload_id,
        workload_type=workload_type,
        dockerfile=dockerfile,
        image_name=f"{RUNNER_IMAGE_REPOSITORY}:{image_hash}",
        sync=True,
    )

    workload_manager().log(
        workload_id=workload_id,
        message="Starting pipeline run.",
    )

    runner_timeout = handle_int_env_var(
        ENV_ZENML_RUNNER_POD_TIMEOUT, default=180
    )

    # Could do this same thing with a step operator, but we need some
    # minor changes to the abstract interface to support that.
    workload_manager().run(
        workload_id=workload_id,
        image=runner_image,
        command=command,
        arguments=arguments,
        environment=environment,
        timeout_in_seconds=runner_timeout,
        sync=wait_for_completion,
    )
    if success_message:
        workload_manager().log(
            workload_id=workload_id,
            message=success_message,
        )



#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Initialization of the Tekton integration for ZenML.

The Tekton integration sub-module powers an alternative to the local
orchestrator. You can enable it by registering the Tekton orchestrator with
the CLI tool.
"""
from typing import List, Type

from zenml.enums import StackComponentType
from zenml.integrations.constants import TEKTON
from zenml.integrations.integration import Integration
from zenml.stack import Flavor

TEKTON_ORCHESTRATOR_FLAVOR = "tekton"


class TektonIntegration(Integration):
    """Definition of Tekton Integration for ZenML."""

    NAME = TEKTON
    REQUIREMENTS = ["kfp>=2.6.0", "kfp-kubernetes>=1.1.0"]
    REQUIREMENTS_IGNORED_ON_UNINSTALL = ["kfp"]

    @classmethod
    def flavors(cls) -> List[Type[Flavor]]:
        """Declare the stack component flavors for the Tekton integration.

        Returns:
            List of stack component flavors for this integration.
        """
        from zenml.integrations.tekton.flavors import TektonOrchestratorFlavor

        return [TektonOrchestratorFlavor]





"""adding exclusive attribute to tags [9e7bf0970266].

Revision ID: 9e7bf0970266
Revises: 2e695a26fe7a
Create Date: 2025-03-03 15:17:49.341208

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "9e7bf0970266"
down_revision = "2e695a26fe7a"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    # Use batch_alter_table for safer schema modifications
    with op.batch_alter_table("tag") as batch_op:
        # First add the column as nullable
        batch_op.add_column(
            sa.Column(
                "exclusive",
                sa.Boolean(),
                nullable=True,
            ),
        )

    # Update existing rows with default value
    op.execute("UPDATE tag SET exclusive = FALSE WHERE exclusive IS NULL")

    # Then alter the column to be non-nullable with a default
    with op.batch_alter_table("tag") as batch_op:
        batch_op.alter_column(
            "exclusive",
            existing_type=sa.Boolean(),
            nullable=False,
            server_default=sa.false(),
        )


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    op.drop_column("tag", "exclusive")



"""Simplify model version links [ec6307720f92].

Revision ID: ec6307720f92
Revises: 0.70.0
Create Date: 2024-11-06 16:16:43.344569

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "ec6307720f92"
down_revision = "0.70.0"
branch_labels = None
depends_on = None


def _migrate_artifact_type() -> None:
    """Migrate the artifact type."""
    meta = sa.MetaData()
    meta.reflect(
        bind=op.get_bind(),
        only=("artifact_version", "model_versions_artifacts"),
    )
    artifact_version_table = sa.Table("artifact_version", meta)
    model_versions_artifacts = sa.Table("model_versions_artifacts", meta)

    connection = op.get_bind()

    query = sa.select(
        model_versions_artifacts.c.artifact_version_id,
        model_versions_artifacts.c.is_model_artifact,
        model_versions_artifacts.c.is_deployment_artifact,
    )
    result = connection.execute(query)

    updated = set()
    updates = []
    for (
        artifact_version_id,
        is_model_artifact,
        is_deployment_artifact,
    ) in result:
        if artifact_version_id in updated:
            # If an artifact was a model artifact in one model version, and a
            # deployment artifact in another model version, we only update it
            # once.
            continue

        if is_model_artifact:
            updated.add(artifact_version_id)
            updates.append(
                {"id_": artifact_version_id, "type": "ModelArtifact"}
            )
        elif is_deployment_artifact:
            updated.add(artifact_version_id)
            updates.append(
                {"id_": artifact_version_id, "type": "ServiceArtifact"}
            )

    if updates:
        connection.execute(
            sa.update(artifact_version_table).where(
                artifact_version_table.c.id == sa.bindparam("id_")
            ),
            updates,
        )


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    # ### commands auto generated by Alembic - please adjust! ###

    _migrate_artifact_type()

    with op.batch_alter_table(
        "model_versions_artifacts", schema=None
    ) as batch_op:
        batch_op.drop_constraint(
            "fk_model_versions_artifacts_workspace_id_workspace",
            type_="foreignkey",
        )
        batch_op.drop_constraint(
            "fk_model_versions_artifacts_user_id_user", type_="foreignkey"
        )
        batch_op.drop_constraint(
            "fk_model_versions_artifacts_model_id_model", type_="foreignkey"
        )
        batch_op.drop_column("user_id")
        batch_op.drop_column("model_id")
        batch_op.drop_column("is_deployment_artifact")
        batch_op.drop_column("workspace_id")
        batch_op.drop_column("is_model_artifact")

    with op.batch_alter_table("model_versions_runs", schema=None) as batch_op:
        batch_op.drop_constraint(
            "fk_model_versions_runs_model_id_model", type_="foreignkey"
        )
        batch_op.drop_constraint(
            "fk_model_versions_runs_workspace_id_workspace", type_="foreignkey"
        )
        batch_op.drop_constraint(
            "fk_model_versions_runs_user_id_user", type_="foreignkey"
        )
        batch_op.drop_column("model_id")
        batch_op.drop_column("workspace_id")
        batch_op.drop_column("user_id")

    # ### end Alembic commands ###


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision.

    Raises:
        NotImplementedError: Downgrade is not supported for this migration.
    """
    raise NotImplementedError("Downgrade is not supported for this migration.")



#  Copyright (c) ZenML GmbH 2023. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Implementation of ZenML's pydantic materializer."""

import hashlib
import json
import os
from typing import TYPE_CHECKING, Any, ClassVar, Dict, Optional, Tuple, Type

from pydantic import BaseModel

from zenml.enums import ArtifactType, VisualizationType
from zenml.materializers.base_materializer import BaseMaterializer
from zenml.utils import yaml_utils

if TYPE_CHECKING:
    from zenml.metadata.metadata_types import MetadataType

DEFAULT_FILENAME = "data.json"


class PydanticMaterializer(BaseMaterializer):
    """Handle Pydantic BaseModel objects."""

    ASSOCIATED_ARTIFACT_TYPE: ClassVar[ArtifactType] = ArtifactType.DATA
    ASSOCIATED_TYPES: ClassVar[Tuple[Type[Any], ...]] = (BaseModel,)

    def load(self, data_type: Type[BaseModel]) -> Any:
        """Reads BaseModel from JSON.

        Args:
            data_type: The type of the data to read.

        Returns:
            The data read.
        """
        data_path = os.path.join(self.uri, DEFAULT_FILENAME)
        contents = yaml_utils.read_json(data_path)
        return data_type.model_validate_json(contents)

    def save(self, data: BaseModel) -> None:
        """Serialize a BaseModel to JSON.

        Args:
            data: The data to store.
        """
        data_path = os.path.join(self.uri, DEFAULT_FILENAME)
        yaml_utils.write_json(data_path, data.model_dump_json())

    def extract_metadata(self, data: BaseModel) -> Dict[str, "MetadataType"]:
        """Extract metadata from the given BaseModel object.

        Args:
            data: The BaseModel object to extract metadata from.

        Returns:
            The extracted metadata as a dictionary.
        """
        return {"schema": data.model_json_schema()}

    def compute_content_hash(self, data: BaseModel) -> Optional[str]:
        """Compute the content hash of the given data.

        Args:
            data: The data to compute the content hash of.

        Returns:
            The content hash of the given data.
        """
        hash_ = hashlib.md5(usedforsecurity=False)
        hash_.update(self.__class__.__name__.encode())

        json_data = data.model_dump(mode="json")
        hash_.update(json.dumps(json_data, sort_keys=True).encode())
        return hash_.hexdigest()

    def save_visualizations(self, data: Any) -> Dict[str, "VisualizationType"]:
        """Save visualizations for the given data.

        Args:
            data: The data to save visualizations for.

        Returns:
            A dictionary of visualization URIs and their types.
        """
        data_path = os.path.join(self.uri, DEFAULT_FILENAME)
        return {data_path.replace("\\", "/"): VisualizationType.JSON}



#  Copyright (c) ZenML GmbH 2025. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Dynamic pipeline execution utilities."""

import time
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Generic,
    Optional,
    Sequence,
    Type,
    TypeVar,
    overload,
)
from uuid import UUID

from zenml.client import Client
from zenml.enums import ExecutionStatus, RunWaitConditionType
from zenml.logger import get_logger
from zenml.models import (
    ArtifactVersionResponse,
    StepRunResponse,
)

if TYPE_CHECKING:
    from zenml.execution.pipeline.dynamic.outputs import (
        AnyStepFuture,
        OutputArtifact,
        StepRunOutputs,
    )

logger = get_logger(__name__)


T = TypeVar("T")


class _Unmapped(Generic[T]):
    """Wrapper class for inputs that should not be mapped over."""

    def __init__(self, value: T):
        """Initialize the wrapper.

        Args:
            value: The value to wrap.
        """
        self.value = value


def unmapped(value: T) -> _Unmapped[T]:
    """Helper function to pass an input without mapping over it.

    Wrap any step input with this function and then pass it to `step.map(...)`
    to pass the full value to all steps.

    Args:
        value: The value to wrap.

    Returns:
        The wrapped value.
    """
    return _Unmapped(value)


@overload
def wait(
    schema: Type[T],
    type: RunWaitConditionType = RunWaitConditionType.EXTERNAL_INPUT,
    timeout: int = 600,
    poll_interval: int = 5,
    question: Optional[str] = None,
    metadata: Optional[Dict[str, Any]] = None,
    after: Optional[Sequence["AnyStepFuture"]] = None,
    name: Optional[str] = None,
) -> T: ...


@overload
def wait(
    schema: object = None,
    type: RunWaitConditionType = RunWaitConditionType.EXTERNAL_INPUT,
    timeout: int = 600,
    poll_interval: int = 5,
    question: Optional[str] = None,
    metadata: Optional[Dict[str, Any]] = None,
    after: Optional[Sequence["AnyStepFuture"]] = None,
    name: Optional[str] = None,
) -> Any: ...


def wait(
    schema: Optional[Any] = None,
    type: RunWaitConditionType = RunWaitConditionType.EXTERNAL_INPUT,
    timeout: int = 600,
    poll_interval: int = 5,
    question: Optional[str] = None,
    metadata: Optional[Dict[str, Any]] = None,
    after: Optional[Sequence["AnyStepFuture"]] = None,
    name: Optional[str] = None,
) -> Any:
    """Pause dynamic execution on an external wait condition.

    Args:
        schema: Optional expected output type for the resolved result.
        type: Wait condition type.
        timeout: Maximum time in seconds to poll before pausing.
        poll_interval: Poll interval in seconds.
        question: Optional question shown to external actors.
        metadata: Optional metadata attached to the condition.
        after: Optional upstream futures that must finish before waiting.
        name: Optional deterministic wait condition name.

    Raises:
        RuntimeError: If called outside of dynamic pipeline execution.

    Returns:
        The resolved wait condition value.
    """
    from zenml.execution.pipeline.dynamic.run_context import (
        DynamicPipelineRunContext,
    )

    context = DynamicPipelineRunContext.get()
    if not context:
        raise RuntimeError(
            "`zenml.wait(...)` can only be used inside dynamic pipelines."
        )

    return context.runner.wait(
        schema=schema,
        type=type,
        timeout=timeout,
        poll_interval=poll_interval,
        question=question,
        after=after,
        metadata=metadata,
        name=name,
    )


def wait_for_step_run_to_finish(step_run_id: UUID) -> "StepRunResponse":
    """Wait until a step run is finished.

    Args:
        step_run_id: The ID of the step run.

    Returns:
        The finished step run.
    """
    sleep_interval = 1
    max_sleep_interval = 64

    while True:
        step_run = Client().zen_store.get_run_step(step_run_id)

        if step_run.status != ExecutionStatus.RUNNING:
            return step_run

        logger.debug(
            "Waiting for step run with ID %s to finish (current status: %s)",
            step_run_id,
            step_run.status,
        )
        time.sleep(sleep_interval)
        if sleep_interval < max_sleep_interval:
            sleep_interval *= 2


def get_latest_step_run(
    pipeline_run_id: UUID, invocation_id: str, hydrate: bool = False
) -> "StepRunResponse":
    """Get the latest step run for a step.

    Args:
        pipeline_run_id: The ID of the pipeline run.
        invocation_id: The invocation ID of the step.
        hydrate: Whether to hydrate the step run.

    Raises:
        RuntimeError: If no step run exists for the given invocation ID.

    Returns:
        The latest step run.
    """
    step_runs = Client().list_run_steps(
        pipeline_run_id=pipeline_run_id,
        name=invocation_id,
        exclude_retried=True,
        size=1,
        hydrate=hydrate,
    )

    if not step_runs:
        raise RuntimeError(
            f"Step `{invocation_id}` not found in pipeline run "
            f"`{pipeline_run_id}`."
        )

    return step_runs.items[0]


def wait_for_step_to_finish(
    pipeline_run_id: UUID, step_name: str
) -> "StepRunResponse":
    """Wait until a step is finished.

    Args:
        pipeline_run_id: The ID of the pipeline run.
        step_name: The name of the step.

    Returns:
        The finished step run.
    """
    sleep_interval = 1
    max_sleep_interval = 64

    while True:
        step_run = get_latest_step_run(
            pipeline_run_id, step_name, hydrate=False
        )
        # If a step is in `retrying` status, another step run will be
        # created and we will try to pick it up in the next iteration.
        if step_run.status not in {
            ExecutionStatus.RUNNING,
            ExecutionStatus.RETRYING,
        }:
            return step_run

        logger.debug(
            "Waiting for step `%s` to finish (current status: %s)",
            step_name,
            step_run.status,
        )

        time.sleep(sleep_interval)
        if sleep_interval < max_sleep_interval:
            sleep_interval *= 2


def load_step_run_outputs(step_run_id: UUID) -> "StepRunOutputs":
    """Load the outputs of a step run.

    Args:
        step_run_id: The ID of the step run.

    Returns:
        The outputs of the step run.
    """
    from zenml.execution.pipeline.dynamic.outputs import OutputArtifact

    step_run = Client().zen_store.get_run_step(step_run_id)

    def _convert_output_artifact(
        output_name: str, artifact: "ArtifactVersionResponse"
    ) -> "OutputArtifact":
        return OutputArtifact(
            output_name=output_name,
            step_name=step_run.name,
            **artifact.model_dump(),
        )

    output_artifacts = step_run.regular_outputs
    if len(output_artifacts) == 0:
        return None
    elif len(output_artifacts) == 1:
        name, artifact = next(iter(output_artifacts.items()))
        return _convert_output_artifact(output_name=name, artifact=artifact)
    else:
        # Make sure we return them in the same order as they're defined in the
        # step configuration, as we don't enforce any ordering in the DB.
        return tuple(
            _convert_output_artifact(
                output_name=name, artifact=output_artifacts[name]
            )
            for name in step_run.config.outputs.keys()
        )



#  Copyright (c) ZenML GmbH 2023. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Models used by the Facets integration."""

from typing import Dict, List, Union

import pandas as pd
from pydantic import BaseModel, ConfigDict


class FacetsComparison(BaseModel):
    """Facets comparison model.

    Returning this from any step will automatically visualize the datasets
    statistics using Facets.

    Attributes:
        datasets: List of datasets to compare. Should be in the format
            `[{"name": "dataset_name", "table": pd.DataFrame}, ...]`.
    """

    datasets: List[Dict[str, Union[str, pd.DataFrame]]]
    model_config = ConfigDict(arbitrary_types_allowed=True)



#  Copyright (c) ZenML GmbH 2026. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""SQLModel implementation of run wait condition tables."""

import json
from datetime import datetime
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Sequence
from uuid import UUID

from sqlalchemy import TEXT, CheckConstraint, Column, UniqueConstraint
from sqlalchemy.orm import joinedload
from sqlalchemy.sql.base import ExecutableOption
from sqlmodel import Field, Relationship

from zenml.enums import (
    MetadataResourceTypes,
    RunWaitConditionStatus,
    RunWaitConditionType,
)
from zenml.models import (
    RunWaitConditionRequest,
    RunWaitConditionResponse,
    RunWaitConditionResponseBody,
    RunWaitConditionResponseMetadata,
    RunWaitConditionResponseResources,
)
from zenml.zen_stores.schemas.base_schemas import BaseSchema
from zenml.zen_stores.schemas.project_schemas import ProjectSchema
from zenml.zen_stores.schemas.schema_utils import (
    build_foreign_key_field,
    build_index,
)
from zenml.zen_stores.schemas.utils import RunMetadataInterface, jl_arg

if TYPE_CHECKING:
    from zenml.zen_stores.schemas.pipeline_run_schemas import PipelineRunSchema
    from zenml.zen_stores.schemas.run_metadata_schemas import RunMetadataSchema
    from zenml.zen_stores.schemas.user_schemas import UserSchema


class RunWaitConditionSchema(BaseSchema, RunMetadataInterface, table=True):
    """SQLModel schema for persisted run wait conditions."""

    __tablename__ = "run_wait_condition"
    __table_args__ = (
        UniqueConstraint(
            "run_id",
            "name",
            name="unique_wait_condition_name_per_run",
        ),
        CheckConstraint(
            "(status != 'resolved') OR (resolution IS NOT NULL)",
            name="ck_run_wait_condition_resolved_requires_resolution",
        ),
        build_index(
            table_name=__tablename__,
            column_names=["run_id", "status"],
        ),
        build_index(
            table_name=__tablename__,
            column_names=["project_id", "created"],
        ),
    )

    run_id: UUID = build_foreign_key_field(
        source=__tablename__,
        target="pipeline_run",
        source_column="run_id",
        target_column="id",
        ondelete="CASCADE",
        nullable=False,
    )
    project_id: UUID = build_foreign_key_field(
        source=__tablename__,
        target=ProjectSchema.__tablename__,
        source_column="project_id",
        target_column="id",
        ondelete="CASCADE",
        nullable=False,
    )
    user_id: Optional[UUID] = build_foreign_key_field(
        source=__tablename__,
        target="user",
        source_column="user_id",
        target_column="id",
        ondelete="SET NULL",
        nullable=True,
    )
    resolved_by_user_id: Optional[UUID] = build_foreign_key_field(
        source=__tablename__,
        target="user",
        source_column="resolved_by_user_id",
        target_column="id",
        ondelete="SET NULL",
        nullable=True,
    )
    type: str = Field(nullable=False)
    status: str = Field(nullable=False)
    name: str = Field(nullable=False)
    question: Optional[str] = Field(default=None, nullable=True)
    data_schema_json: Optional[str] = Field(
        default=None, sa_column=Column(TEXT, nullable=True)
    )
    resolution: Optional[str] = Field(default=None, nullable=True)
    result_json: Optional[str] = Field(default=None, sa_column=Column(TEXT))
    last_polled_at: Optional[datetime] = Field(nullable=True, default=None)
    poller_instance_id: Optional[str] = Field(default=None, nullable=True)
    poller_lease_expires_at: Optional[datetime] = Field(
        nullable=True, default=None
    )
    resolved_at: Optional[datetime] = Field(nullable=True, default=None)
    run: "PipelineRunSchema" = Relationship(back_populates="wait_conditions")
    user: Optional["UserSchema"] = Relationship(
        sa_relationship_kwargs={
            "foreign_keys": "RunWaitConditionSchema.user_id",
        }
    )
    resolved_by_user: Optional["UserSchema"] = Relationship(
        sa_relationship_kwargs={
            "foreign_keys": "RunWaitConditionSchema.resolved_by_user_id",
        }
    )
    run_metadata: List["RunMetadataSchema"] = Relationship(
        sa_relationship_kwargs=dict(
            secondary="run_metadata_resource",
            primaryjoin=f"and_(foreign(RunMetadataResourceSchema.resource_type)=='{MetadataResourceTypes.WAIT_CONDITION.value}', foreign(RunMetadataResourceSchema.resource_id)==RunWaitConditionSchema.id)",
            secondaryjoin="RunMetadataSchema.id==foreign(RunMetadataResourceSchema.run_metadata_id)",
            overlaps="run_metadata",
        ),
    )

    @classmethod
    def get_query_options(
        cls,
        include_metadata: bool = False,
        include_resources: bool = False,
        **kwargs: Any,
    ) -> Sequence[ExecutableOption]:
        """Get query options for converting schema rows to models.

        Args:
            include_metadata: Whether metadata will be included in responses.
            include_resources: Whether resources will be included in responses.
            **kwargs: Additional unused keyword arguments.

        Returns:
            SQLAlchemy query options.
        """
        options: List[ExecutableOption] = [joinedload(jl_arg(cls.run))]
        if include_resources:
            options.extend(
                [
                    joinedload(jl_arg(cls.user)),
                    joinedload(jl_arg(cls.resolved_by_user)),
                ]
            )
        return options

    @classmethod
    def from_request(
        cls, request: RunWaitConditionRequest
    ) -> "RunWaitConditionSchema":
        """Create a schema object from a wait condition create request.

        Args:
            request: Wait condition creation request.

        Returns:
            The persisted wait condition schema object.
        """
        return cls(
            run_id=request.run,
            project_id=request.project,
            user_id=request.user,
            name=request.name,
            type=request.type.value,
            status=RunWaitConditionStatus.PENDING.value,
            question=request.question,
            data_schema_json=(
                json.dumps(request.data_schema)
                if request.data_schema is not None
                else None
            ),
        )

    def to_model(
        self,
        include_metadata: bool = False,
        include_resources: bool = False,
        **kwargs: Any,
    ) -> RunWaitConditionResponse:
        """Convert the schema row to a response model.

        Args:
            include_metadata: Whether metadata should be included.
            include_resources: Whether resources should be included.
            **kwargs: Additional unused keyword arguments.

        Returns:
            The wait condition response model.
        """
        data_schema: Optional[Dict[str, Any]] = None
        if self.data_schema_json:
            data_schema = json.loads(self.data_schema_json)

        result: Optional[Any] = None
        if self.result_json:
            result = json.loads(self.result_json)

        body = RunWaitConditionResponseBody(
            user_id=self.user_id,
            project_id=self.run.project_id,
            created=self.created,
            updated=self.updated,
            type=RunWaitConditionType(self.type),
            status=RunWaitConditionStatus(self.status),
            last_polled_at=self.last_polled_at,
            poller_instance_id=self.poller_instance_id,
            poller_lease_expires_at=self.poller_lease_expires_at,
            resolved_at=self.resolved_at,
            resolved_by_user_id=self.resolved_by_user_id,
        )

        metadata = None
        if include_metadata:
            metadata = RunWaitConditionResponseMetadata(
                question=self.question,
                run_metadata=self.fetch_metadata(),
                data_schema=data_schema,
                resolution=self.resolution,
                result=result,
            )

        resources = None
        if include_resources:
            resources = RunWaitConditionResponseResources(
                user=self.user.to_model() if self.user else None,
                run=self.run.to_model(
                    include_metadata=False, include_resources=False
                ),
            )

        return RunWaitConditionResponse(
            id=self.id,
            name=self.name,
            body=body,
            metadata=metadata,
            resources=resources,
        )



#  Copyright (c) ZenML GmbH 2025. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Dynamic pipeline execution outputs."""

from abc import ABC, abstractmethod
from concurrent.futures import Future
from typing import Any, Iterator, List, Optional, Tuple, Union, overload
from uuid import UUID

from zenml.logger import get_logger
from zenml.models import ArtifactVersionResponse, StepRunResponse
from zenml.utils import exception_utils

logger = get_logger(__name__)


class OutputArtifact(ArtifactVersionResponse):
    """Dynamic step run output artifact."""

    output_name: str
    step_name: str
    chunk_index: Optional[int] = None
    chunk_size: Optional[int] = None

    def chunk(self, index: int) -> "OutputArtifact":
        """Get a chunk of the output artifact.

        Args:
            index: The index of the chunk.

        Raises:
            ValueError: If the output artifact can not be chunked or the index
                is out of range.

        Returns:
            The artifact chunk.
        """
        if not self.item_count:
            raise ValueError(
                f"Output artifact `{self.output_name}` of step "
                f"`{self.step_name}` can not be chunked."
            )

        if index < 0 or index >= self.item_count:
            raise ValueError(
                f"Chunk index `{index}` out of range for output artifact "
                f"`{self.output_name}` of step `{self.step_name}`."
            )

        if self.chunk_index is not None and self.chunk_index != index:
            raise ValueError(
                f"Output artifact `{self.output_name}` of step "
                f"`{self.step_name}` is already referring to a "
                "different chunk."
            )

        return self.model_copy(update={"chunk_index": index, "chunk_size": 1})


StepRunOutputs = Union[None, OutputArtifact, Tuple[OutputArtifact, ...]]


class BaseFuture(ABC):
    """Base future."""

    @abstractmethod
    def running(self) -> bool:
        """Check if the future is running.

        Returns:
            True if the future is running, False otherwise.
        """

    @abstractmethod
    def result(self) -> Any:
        """Get the result of the future.

        Returns:
            The result of the future.
        """


class _InlineStepFuture(BaseFuture):
    """Future for an inline step run."""

    def __init__(
        self, wrapped: Future["StepRunResponse"], invocation_id: str
    ) -> None:
        """Initialize the inline step run future.

        Args:
            wrapped: The wrapped future object.
            invocation_id: The invocation ID of the step run.
        """
        self._wrapped = wrapped
        self.invocation_id = invocation_id

    def running(self) -> bool:
        """Check if the step run future is running.

        Returns:
            True if the step run future is running, False otherwise.
        """
        return not self._wrapped.done()

    def result(self) -> "StepRunResponse":
        """Get the result of the step run future.

        Returns:
            The result of the step run future.
        """
        return self._wrapped.result()


class _IsolatedStepFuture(BaseFuture):
    """Future for an inline step run."""

    def __init__(
        self,
        pipeline_run_id: UUID,
        invocation_id: str,
        wrapped: Optional[Future["StepRunResponse"]] = None,
    ) -> None:
        """Initialize the step run future.

        Args:
            pipeline_run_id: The ID of the pipeline run.
            invocation_id: The invocation ID of the step run.
            wrapped: Optional future to wait for that submits the step run.
        """
        self._wrapped = wrapped
        self.pipeline_run_id = pipeline_run_id
        self.invocation_id = invocation_id

    def running(self) -> bool:
        """Check if the isolated step future is running.

        Returns:
            True if the isolated step future is running, False otherwise.
        """
        from zenml.execution.pipeline.dynamic.utils import get_latest_step_run

        if self._wrapped and not self._wrapped.done():
            return True

        step_run = get_latest_step_run(
            self.pipeline_run_id, self.invocation_id, hydrate=False
        )

        return not step_run.status.is_finished

    def result(self) -> "StepRunResponse":
        """Get the result of the step future.

        # noqa: DAR401
        Raises:
            BaseException: Any exception that happened while waiting for the
                step to finish.

        Returns:
            The result of the step future.
        """
        from zenml.execution.pipeline.dynamic.utils import (
            wait_for_step_to_finish,
        )

        if self._wrapped:
            # We first wait until the step run is submitted and only then
            # start monitoring the actual step.
            self._wrapped.result()

        step_run = wait_for_step_to_finish(
            pipeline_run_id=self.pipeline_run_id, step_name=self.invocation_id
        )

        if step_run.status.is_failed:
            raise exception_utils.reconstruct_exception(
                exception_info=step_run.exception_info,
                fallback_message=(
                    f"Step `{self.invocation_id}` failed with "
                    f"status `{step_run.status}`."
                ),
            )

        return step_run


class BaseStepFuture(BaseFuture):
    """Base step future."""

    def __init__(
        self,
        wrapped: Union[_InlineStepFuture, _IsolatedStepFuture],
        **kwargs: Any,
    ) -> None:
        """Initialize the dynamic step run future.

        Args:
            wrapped: The wrapped future object.
            **kwargs: Additional keyword arguments.
        """
        self._wrapped = wrapped

    @property
    def invocation_id(self) -> str:
        """The step run invocation ID.

        Returns:
            The step run invocation ID.
        """
        return self._wrapped.invocation_id

    def running(self) -> bool:
        """Check if the step run future is running.

        Returns:
            True if the step run future is running, False otherwise.
        """
        return self._wrapped.running()


class ArtifactFuture(BaseStepFuture):
    """Future for a step run output artifact."""

    def __init__(
        self,
        wrapped: Union[_InlineStepFuture, _IsolatedStepFuture],
        index: int,
    ) -> None:
        """Initialize the future.

        Args:
            wrapped: The wrapped future object.
            index: The index of the output artifact.
        """
        super().__init__(wrapped=wrapped)
        self._index = index

    def result(self) -> OutputArtifact:
        """Get the output artifact this future represents.

        Raises:
            RuntimeError: If the future returned an invalid output.

        Returns:
            The output artifact.
        """
        step_run = self._wrapped.result()
        from zenml.execution.pipeline.dynamic.utils import (
            load_step_run_outputs,
        )

        result = load_step_run_outputs(step_run.id)

        if isinstance(result, OutputArtifact):
            return result
        elif isinstance(result, tuple):
            return result[self._index]
        else:
            raise RuntimeError(
                f"Step {self.invocation_id} returned an invalid output: "
                f"{result}."
            )

    def load(self, disable_cache: bool = False) -> Any:
        """Load the step run output artifact data.

        Args:
            disable_cache: Whether to disable the artifact cache.

        Returns:
            The step run output artifact data.
        """
        return self.result().load(disable_cache=disable_cache)

    def chunk(self, index: int) -> "OutputArtifact":
        """Get a chunk of the output artifact.

        This method will wait for the future to complete and then return the
        artifact chunk.

        Args:
            index: The index of the chunk.

        Returns:
            The artifact chunk.
        """
        return self.result().chunk(index=index)


class StepFuture(BaseStepFuture):
    """Future for a step run output."""

    def __init__(
        self,
        wrapped: Union[_InlineStepFuture, _IsolatedStepFuture],
        output_keys: List[str],
    ) -> None:
        """Initialize the future.

        Args:
            wrapped: The wrapped future object.
            output_keys: The output keys of the step run.
        """
        super().__init__(wrapped=wrapped)
        self._output_keys = output_keys

    def get_artifact(self, key: str) -> ArtifactFuture:
        """Get an artifact future by key.

        Args:
            key: The key of the artifact future.

        Raises:
            KeyError: If no artifact for the given name exists.

        Returns:
            The artifact future.
        """
        if key not in self._output_keys:
            raise KeyError(
                f"Step run {self.invocation_id} does not have an output with "
                f"the name: {key}."
            )

        return ArtifactFuture(
            wrapped=self._wrapped,
            index=self._output_keys.index(key),
        )

    def wait(self) -> None:
        """Wait for the step to finish."""
        self._wrapped.result()

    def artifacts(self) -> StepRunOutputs:
        """Get the step run output artifacts.

        Returns:
            The step run output artifacts.
        """
        return self.result()

    def result(self) -> StepRunOutputs:
        """Get the step run outputs this future represents.

        Returns:
            The step run outputs.
        """
        from zenml.execution.pipeline.dynamic.utils import (
            load_step_run_outputs,
        )

        step_run = self._wrapped.result()
        return load_step_run_outputs(step_run.id)

    def load(self, disable_cache: bool = False) -> Any:
        """Get the step run output artifact data.

        Args:
            disable_cache: Whether to disable the artifact cache.

        Raises:
            ValueError: If the step run output is invalid.

        Returns:
            The step run output artifact data.
        """
        result = self.artifacts()

        if result is None:
            return None
        elif isinstance(result, ArtifactVersionResponse):
            return result.load(disable_cache=disable_cache)
        elif isinstance(result, tuple):
            return tuple(
                item.load(disable_cache=disable_cache) for item in result
            )
        else:
            raise ValueError(f"Invalid step run output: {result}")

    @overload
    def __getitem__(self, key: int) -> ArtifactFuture: ...

    @overload
    def __getitem__(self, key: slice) -> Tuple[ArtifactFuture, ...]: ...

    def __getitem__(
        self, key: Union[int, slice]
    ) -> Union[ArtifactFuture, Tuple[ArtifactFuture, ...]]:
        """Get an artifact future.

        Args:
            key: The index or slice of the artifact futures.

        Raises:
            TypeError: If the key is not an integer or slice.

        Returns:
            The artifact futures.
        """
        if isinstance(key, int):
            output_key = self._output_keys[key]

            return ArtifactFuture(
                wrapped=self._wrapped,
                index=self._output_keys.index(output_key),
            )
        elif isinstance(key, slice):
            output_keys = self._output_keys[key]
            return tuple(
                ArtifactFuture(
                    wrapped=self._wrapped,
                    index=self._output_keys.index(output_key),
                )
                for output_key in output_keys
            )
        else:
            raise TypeError(f"Invalid key type: {type(key)}")

    def __iter__(self) -> Any:
        """Iterate over the artifact futures.

        Raises:
            ValueError: If the step does not return any outputs.

        Yields:
            The artifact futures.
        """
        if not self._output_keys:
            raise ValueError(
                f"Step {self.invocation_id} does not return any outputs."
            )

        for index in range(len(self._output_keys)):
            yield ArtifactFuture(
                wrapped=self._wrapped,
                index=index,
            )

    def __len__(self) -> int:
        """Get the number of artifact futures.

        Returns:
            The number of artifact futures.
        """
        return len(self._output_keys)


class MapResultsFuture(BaseFuture):
    """Future that represents the results of a `step.map/product(...)` call."""

    def __init__(self, futures: List[StepFuture]) -> None:
        """Initialize the map results future.

        Args:
            futures: The step run futures.
        """
        self.futures = futures

    def running(self) -> bool:
        """Check if the map results future is running.

        Returns:
            True if the map results future is running, False otherwise.
        """
        return any(future.running() for future in self.futures)

    def result(self) -> List[StepRunOutputs]:
        """Get the step run outputs this future represents.

        Returns:
            The step run outputs.
        """
        return [future.result() for future in self.futures]

    def load(self, disable_cache: bool = False) -> List[Any]:
        """Load the step run output artifacts.

        Args:
            disable_cache: Whether to disable the artifact cache.

        Returns:
            The step run output artifacts.
        """
        return [
            future.load(disable_cache=disable_cache) for future in self.futures
        ]

    def unpack(self) -> Tuple[List[ArtifactFuture], ...]:
        """Unpack the map results future.

        This method can be used to get lists of artifact futures that represent
        the outputs of all the step runs that are part of this map result.

        Example:
        ```python
        from zenml import pipeline, step

        @step
        def create_int_list() -> list[int]:
            return [1, 2]

        @step
        def do_something(a: int) -> Tuple[int, int]:
            return a * 2, a * 3

        @pipeline
        def map_pipeline():
            int_list = create_int_list()
            results = do_something.map(a=int_list)
            double, triple = results.unpack()

            # [future.load() for future in double] will return [2, 4]
            # [future.load() for future in triple] will return [3, 6]
        ```

        Returns:
            The unpacked map results.
        """
        return tuple(map(list, zip(*self.futures)))

    @overload
    def __getitem__(self, key: int) -> StepFuture: ...

    @overload
    def __getitem__(self, key: slice) -> List[StepFuture]: ...

    def __getitem__(
        self, key: Union[int, slice]
    ) -> Union[StepFuture, List[StepFuture]]:
        """Get a step run future.

        Args:
            key: The index or slice of the step run futures.

        Returns:
            The step run futures.
        """
        return self.futures[key]

    def __iter__(self) -> Iterator[StepFuture]:
        """Iterate over the step run futures.

        Yields:
            The step run futures.
        """
        yield from self.futures

    def __len__(self) -> int:
        """Get the number of step run futures.

        Returns:
            The number of step run futures.
        """
        return len(self.futures)


AnyStepFuture = Union[ArtifactFuture, StepFuture, MapResultsFuture]



#  Copyright (c) ZenML GmbH 2023. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Only-load materializer for directories."""

import os
from pathlib import Path
from typing import Any, ClassVar, Tuple, Type

from zenml.enums import ArtifactType
from zenml.io import fileio
from zenml.materializers.base_materializer import BaseMaterializer


class PreexistingDataMaterializer(BaseMaterializer):
    """Materializer to load directories from the artifact store.

    This materializer is very special, since it do not implement save
    logic at all. The save of the data to some URI inside the artifact store
    shall happen outside and is in user's responsibility.

    This materializer solely supports the `register_artifact` function.
    """

    ASSOCIATED_TYPES: ClassVar[Tuple[Type[Any], ...]] = (Path,)
    ASSOCIATED_ARTIFACT_TYPE: ClassVar[ArtifactType] = ArtifactType.DATA
    SKIP_REGISTRATION: ClassVar[bool] = True

    def load(self, data_type: Type[Any]) -> Any:
        """Copy the artifact file(s) to a local temp directory.

        Args:
            data_type: Unused.

        Returns:
            Path to the local directory that contains the artifact files.
        """
        with self.get_temporary_directory(delete_at_exit=False) as temp_dir:
            if fileio.isdir(self.uri):
                self._copy_directory(src=self.uri, dst=temp_dir)
                return Path(temp_dir)
            else:
                dst = os.path.join(temp_dir, os.path.split(self.uri)[-1])
                fileio.copy(src=self.uri, dst=dst)
                return Path(dst)

    def save(self, data: Any) -> None:
        """Store the directory in the artifact store.

        Args:
            data: Path to a local directory to store.

        Raises:
            NotImplementedError: Always
        """
        raise NotImplementedError(
            "`PreexistingDataMaterializer` can only be used in the "
            "context of `register_artifact` function, "
            "which expects the data to be already properly saved in "
            "the Artifact Store, thus `save` logic makes no sense here."
        )

    @staticmethod
    def _copy_directory(src: str, dst: str) -> None:
        """Recursively copy a directory.

        Args:
            src: The directory to copy.
            dst: Where to copy the directory to.
        """
        for src_dir, _, files in fileio.walk(src):
            src_dir_ = str(src_dir)
            dst_dir = str(os.path.join(dst, os.path.relpath(src_dir_, src)))
            fileio.makedirs(dst_dir)

            for file in files:
                file_ = str(file)
                src_file = os.path.join(src_dir_, file_)
                dst_file = os.path.join(dst_dir, file_)
                fileio.copy(src_file, dst_file)



#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""SQL Model Implementations for Flavors."""

import json
from typing import Any, Optional, Sequence
from uuid import UUID

from sqlalchemy import TEXT, Column, UniqueConstraint
from sqlalchemy.orm import joinedload
from sqlalchemy.sql.base import ExecutableOption
from sqlmodel import Field, Relationship

from zenml.enums import StackComponentType
from zenml.models import (
    FlavorResponse,
    FlavorResponseBody,
    FlavorResponseMetadata,
    FlavorResponseResources,
    FlavorUpdate,
)
from zenml.utils.time_utils import utc_now
from zenml.zen_stores.schemas.base_schemas import NamedSchema
from zenml.zen_stores.schemas.schema_utils import build_foreign_key_field
from zenml.zen_stores.schemas.user_schemas import UserSchema
from zenml.zen_stores.schemas.utils import jl_arg


class FlavorSchema(NamedSchema, table=True):
    """SQL Model for flavors.

    Attributes:
        type: The type of the flavor.
        source: The source of the flavor.
        config_schema: The config schema of the flavor.
        integration: The integration associated with the flavor.
    """

    __tablename__ = "flavor"
    __table_args__ = (
        UniqueConstraint(
            "name",
            "type",
            name="unique_flavor_name_and_type",
        ),
    )

    type: str
    source: str
    display_name: Optional[str] = Field(default=None)
    config_schema: str = Field(sa_column=Column(TEXT, nullable=False))
    integration: Optional[str] = Field(default="")
    connector_type: Optional[str]
    connector_resource_type: Optional[str]
    connector_resource_id_attr: Optional[str]

    user_id: Optional[UUID] = build_foreign_key_field(
        source=__tablename__,
        target=UserSchema.__tablename__,
        source_column="user_id",
        target_column="id",
        ondelete="SET NULL",
        nullable=True,
    )
    user: Optional["UserSchema"] = Relationship(back_populates="flavors")

    logo_url: Optional[str] = Field()

    docs_url: Optional[str] = Field()

    sdk_docs_url: Optional[str] = Field()

    is_custom: bool = Field(default=True)

    @classmethod
    def get_query_options(
        cls,
        include_metadata: bool = False,
        include_resources: bool = False,
        **kwargs: Any,
    ) -> Sequence[ExecutableOption]:
        """Get the query options for the schema.

        Args:
            include_metadata: Whether metadata will be included when converting
                the schema to a model.
            include_resources: Whether resources will be included when
                converting the schema to a model.
            **kwargs: Keyword arguments to allow schema specific logic

        Returns:
            A list of query options.
        """
        options = []

        if include_resources:
            options.extend(
                [
                    joinedload(jl_arg(FlavorSchema.user)),
                ]
            )

        return options

    def update(
        self,
        flavor_update: "FlavorUpdate",
    ) -> "FlavorSchema":
        """Update a `FlavorSchema` from a `FlavorUpdate`.

        Args:
            flavor_update: The `FlavorUpdate` from which to update the schema.

        Returns:
            The updated `FlavorSchema`.
        """
        for field, value in flavor_update.model_dump(
            exclude_unset=True, exclude={"user"}
        ).items():
            if field == "config_schema":
                setattr(self, field, json.dumps(value))
            elif field == "type":
                setattr(self, field, value.value)
            else:
                setattr(self, field, value)

        self.updated = utc_now()
        return self

    def to_model(
        self,
        include_metadata: bool = False,
        include_resources: bool = False,
        **kwargs: Any,
    ) -> "FlavorResponse":
        """Converts a flavor schema to a flavor model.

        Args:
            include_metadata: Whether the metadata will be filled.
            include_resources: Whether the resources will be filled.
            **kwargs: Keyword arguments to allow schema specific logic


        Returns:
            The flavor model.
        """
        body = FlavorResponseBody(
            user_id=self.user_id,
            type=StackComponentType(self.type),
            display_name=self.display_name
            or self.name.replace("_", " ").title(),
            integration=self.integration,
            source=self.source,
            logo_url=self.logo_url,
            is_custom=self.is_custom,
            created=self.created,
            updated=self.updated,
        )
        metadata = None
        if include_metadata:
            metadata = FlavorResponseMetadata(
                config_schema=json.loads(self.config_schema),
                connector_type=self.connector_type,
                connector_resource_type=self.connector_resource_type,
                connector_resource_id_attr=self.connector_resource_id_attr,
                docs_url=self.docs_url,
                sdk_docs_url=self.sdk_docs_url,
            )
        resources = None
        if include_resources:
            resources = FlavorResponseResources(
                user=self.user.to_model() if self.user else None,
            )
        return FlavorResponse(
            id=self.id,
            name=self.name,
            body=body,
            metadata=metadata,
            resources=resources,
        )



"""track secrets in db [7b651bf6822e].

Revision ID: 7b651bf6822e
Revises: 1ac1b9c04da1
Create Date: 2023-12-22 20:56:18.131906

"""

from abc import abstractmethod
from datetime import datetime
from typing import Any, Dict, List, Optional
from uuid import UUID

import sqlalchemy as sa
from alembic import op
from pydantic import BaseModel

from zenml.logger import get_logger

logger = get_logger(__name__)

# revision identifiers, used by Alembic.


revision = "7b651bf6822e"
down_revision = "1ac1b9c04da1"
branch_labels = None
depends_on = None


ZENML_SECRET_LABEL = "zenml"
ZENML_SECRET_ID_LABEL = "zenml_secret_id"
ZENML_SECRET_NAME_LABEL = "zenml_secret_name"
ZENML_SECRET_SCOPE_LABEL = "zenml_secret_scope"
ZENML_SECRET_USER_LABEL = "zenml_secret_user"
ZENML_SECRET_WORKSPACE_LABEL = "zenml_secret_workspace"


class ZenMLSecretMetadata(BaseModel):
    """ZenML secret metadata.

    This class reflects the structure of the ZenML secret metadata stored in the
    backend secret (e.g. in secret tags/labels) at the time when the secret
    metadata was migrated to the database. It is roughly aligned to the
    structure of the ZenML SecretResponse model as it was at that time.
    """

    id: UUID
    name: str
    user_id: UUID
    workspace_id: UUID
    created: datetime
    updated: datetime
    scope: str


class BaseSecretsStoreBackend(BaseModel):
    """Base class for accessing and listing ZenML secrets."""

    server_id: UUID
    client: Any = None

    def _get_secret_metadata(
        self,
        secret_id: Optional[UUID] = None,
    ) -> Dict[str, str]:
        """Get a dictionary with metadata that can be used as tags/labels.

        This utility method can be used with Secrets Managers that can
        associate metadata (e.g. tags, labels) with a secret. The metadata can
        be configured alongside each secret.

        NOTE: the ZENML_SECRET_LABEL is always included in the metadata to
        distinguish ZenML secrets from other secrets that might be stored in
        the same backend, as well as to distinguish between different ZenML
        deployments using the same backend. Its value is set to the ZenML
        deployment ID.

        Args:
            secret_id: Optional secret ID to include in the metadata.

        Returns:
            Dictionary with secret metadata information.
        """
        # Always include the main ZenML label to distinguish ZenML secrets
        # from other secrets that might be stored in the same backend and
        # to distinguish between different ZenML deployments using the same
        # backend.
        metadata: Dict[str, str] = {ZENML_SECRET_LABEL: str(self.server_id)}

        # Include the secret ID if provided.
        if secret_id is not None:
            metadata[ZENML_SECRET_ID_LABEL] = str(secret_id)

        return metadata

    def _create_secret_from_metadata(
        self,
        metadata: Dict[str, str],
        created: datetime,
        updated: datetime,
    ) -> ZenMLSecretMetadata:
        """Create a ZenML secret metadata object from metadata stored in the secrets store backend.

        Args:
            metadata: ZenML secret metadata collected from the backend secret
                (e.g. from secret tags/labels).
            created: The secret creation time.
            updated: The secret last updated time.

        Returns:
            The ZenML secret.

        Raises:
            KeyError: If the secret does not have the required metadata, if it
                is not managed by this ZenML instance or if it is linked to a
                user or workspace that no longer exists.
        """
        # Double-check that the secret is managed by this ZenML instance.
        if metadata.get(ZENML_SECRET_LABEL) != str(self.server_id):
            raise KeyError("Secret is not managed by this ZenML instance")

        # Recover the ZenML secret fields from the input secret metadata.
        try:
            secret_id = UUID(metadata[ZENML_SECRET_ID_LABEL])
            name = metadata[ZENML_SECRET_NAME_LABEL]
            scope = metadata[ZENML_SECRET_SCOPE_LABEL]
            workspace_id = UUID(metadata[ZENML_SECRET_WORKSPACE_LABEL])
            user_id = UUID(metadata[ZENML_SECRET_USER_LABEL])
        except KeyError as e:
            raise KeyError(
                f"Secret could not be retrieved: missing required metadata: {e}"
            )

        secret_model = ZenMLSecretMetadata(
            id=secret_id,
            name=name,
            workspace_id=workspace_id,
            user_id=user_id,
            created=created,
            updated=updated,
            scope=scope,
        )

        return secret_model

    @abstractmethod
    def list_secrets(
        self,
    ) -> List[ZenMLSecretMetadata]:
        """List all ZenML secrets in the secrets store backend.

        Returns:
            A list of all secrets.
        """


class AWSSecretsStoreBackend(BaseSecretsStoreBackend):
    """AWS ZenML secrets store backend."""

    @staticmethod
    def _get_aws_secret_filters(
        metadata: Dict[str, str],
    ) -> List[Dict[str, str]]:
        """Convert ZenML secret metadata to AWS secret filters.

        Args:
            metadata: The ZenML secret metadata.

        Returns:
            The AWS secret filters.
        """
        aws_filters: List[Dict[str, Any]] = []
        for k, v in metadata.items():
            aws_filters.append(
                {
                    "Key": "tag-key",
                    "Values": [
                        k,
                    ],
                }
            )
            aws_filters.append(
                {
                    "Key": "tag-value",
                    "Values": [
                        str(v),
                    ],
                }
            )

        return aws_filters

    def list_secrets(
        self,
    ) -> List[ZenMLSecretMetadata]:
        """List all ZenML secrets in the AWS secrets store backend.

        Returns:
            A list of all secrets.

        Raises:
            RuntimeError: If the AWS Secrets Manager API returns an unexpected
                error.
        """
        from botocore.exceptions import ClientError

        # The metadata will always contain at least the filter criteria
        # required to exclude everything but AWS secrets that belong to the
        # current ZenML deployment.
        metadata = self._get_secret_metadata()
        aws_filters = self._get_aws_secret_filters(metadata)

        results: List[ZenMLSecretMetadata] = []

        try:
            # AWS Secrets Manager API pagination is wrapped around the
            # `list_secrets` method call. We use it because we need to fetch all
            # secrets matching the (partial) filter that we set up. Note that
            # the pagination used here has nothing to do with the pagination
            # that we do for the method caller.
            paginator = self.client.get_paginator("list_secrets")
            pages = paginator.paginate(
                Filters=aws_filters,
                PaginationConfig={
                    "PageSize": 100,
                },
            )

            for page in pages:
                for secret in page["SecretList"]:
                    try:
                        # Convert the AWS secret tags to a metadata dictionary.
                        unpacked_metadata: Dict[str, str] = {
                            tag["Key"]: tag["Value"] for tag in secret["Tags"]
                        }

                        secret_model = self._create_secret_from_metadata(
                            metadata=unpacked_metadata,
                            created=secret["CreatedDate"],
                            updated=secret["LastChangedDate"],
                        )
                    except KeyError:
                        # The _convert_aws_secret method raises a KeyError
                        # if the secret is not well formed. Here we pretend that
                        # the secret does not exist.
                        continue

                    results.append(secret_model)
        except ClientError as e:
            raise RuntimeError(f"Error listing AWS secrets: {e}")

        return results


ZENML_GCP_DATE_FORMAT_STRING = "%Y-%m-%d-%H-%M-%S"
ZENML_GCP_SECRET_CREATED_KEY = "zenml-secret-created"
ZENML_GCP_SECRET_UPDATED_KEY = "zenml-secret-updated"


class GCPSecretsStoreBackend(BaseSecretsStoreBackend):
    """GCP ZenML secrets store backend."""

    project_id: str

    @property
    def parent_name(self) -> str:
        """Construct the GCP parent path to the secret manager.

        Returns:
            The parent path to the secret manager
        """
        return f"projects/{self.project_id}"

    def _convert_gcp_secret(
        self,
        labels: Dict[str, str],
    ) -> ZenMLSecretMetadata:
        """Create a ZenML secret model from data stored in an GCP secret.

        If the GCP secret cannot be converted, the method acts as if the
        secret does not exist and raises a KeyError.

        Args:
            labels: The GCP secret labels.

        Returns:
            The ZenML secret model.

        Raises:
            KeyError: if the GCP secret cannot be converted.
        """
        # Recover the ZenML secret metadata from the AWS secret tags.

        # The GCP secret labels do not really behave like a dictionary: when
        # a key is not found, it does not raise a KeyError, but instead
        # returns an empty string. That's why we make this conversion.
        label_dict = dict(labels)

        try:
            created = datetime.strptime(
                label_dict[ZENML_GCP_SECRET_CREATED_KEY],
                ZENML_GCP_DATE_FORMAT_STRING,
            )
            updated = datetime.strptime(
                label_dict[ZENML_GCP_SECRET_UPDATED_KEY],
                ZENML_GCP_DATE_FORMAT_STRING,
            )
        except KeyError as e:
            raise KeyError(
                f"Invalid GCP secret: missing required tag '{e}'"
            ) from e

        return self._create_secret_from_metadata(
            metadata=label_dict,
            created=created,
            updated=updated,
        )

    def list_secrets(self) -> List[ZenMLSecretMetadata]:
        """List all secrets.

        Returns:
            A list of all secrets.

        Raises:
            RuntimeError: if the GCP Secrets Manager API returns an unexpected
                error.
        """
        from google.cloud.secretmanager import SecretManagerServiceClient

        assert isinstance(self.client, SecretManagerServiceClient)

        try:
            # get all the secrets and their labels from GCP
            secrets = []
            for secret in self.client.list_secrets(
                request={
                    "parent": self.parent_name,
                    "filter": "",
                }
            ):
                try:
                    secrets.append(self._convert_gcp_secret(secret.labels))
                except KeyError:
                    # keep going / ignore if this secret version doesn't exist
                    # or isn't a ZenML secret
                    continue
        except Exception as e:
            raise RuntimeError(f"Error listing GCP secrets: {e}") from e

        return secrets


ZENML_AZURE_SECRET_CREATED_KEY = "zenml_secret_created"
ZENML_AZURE_SECRET_UPDATED_KEY = "zenml_secret_updated"


class AzureSecretsStoreBackend(BaseSecretsStoreBackend):
    """Azure ZenML secrets store backend."""

    def _convert_azure_secret(
        self,
        tags: Dict[str, str],
    ) -> ZenMLSecretMetadata:
        """Create a ZenML secret model from data stored in an Azure secret.

        If the Azure secret cannot be converted, the method acts as if the
        secret does not exist and raises a KeyError.

        Args:
            tags: The Azure secret tags.

        Returns:
            The ZenML secret.

        Raises:
            KeyError: if the Azure secret cannot be converted.
        """
        try:
            created = datetime.fromisoformat(
                tags[ZENML_AZURE_SECRET_CREATED_KEY],
            )
            updated = datetime.fromisoformat(
                tags[ZENML_AZURE_SECRET_UPDATED_KEY],
            )
        except KeyError as e:
            raise KeyError(
                f"Secret could not be retrieved: missing required metadata: {e}"
            )

        return self._create_secret_from_metadata(
            metadata=tags,
            created=created,
            updated=updated,
        )

    def list_secrets(self) -> List[ZenMLSecretMetadata]:
        """List all secrets.

        Returns:
            A list of all secrets.

        Raises:
            RuntimeError: If the Azure Key Vault API returns an unexpected
                error.
        """
        from azure.core.exceptions import HttpResponseError
        from azure.keyvault.secrets import SecretClient

        assert isinstance(self.client, SecretClient)

        results: List[ZenMLSecretMetadata] = []

        try:
            all_secrets = self.client.list_properties_of_secrets()
            for secret_property in all_secrets:
                try:
                    # NOTE: we do not include the secret values in the
                    # response. We would need a separate API call to fetch
                    # them for each secret, which would be very inefficient
                    # anyway.
                    assert secret_property.tags is not None
                    secret_model = self._convert_azure_secret(
                        tags=secret_property.tags,
                    )
                except KeyError:
                    # The _convert_azure_secret method raises a KeyError
                    # if the secret is tied to a workspace or user that no
                    # longer exists or if it is otherwise not valid. Here we
                    # pretend that the secret does not exist.
                    continue

                results.append(secret_model)
        except HttpResponseError as e:
            raise RuntimeError(f"Error listing Azure Key Vault secrets: {e}")

        return results


HVAC_ZENML_SECRET_NAME_PREFIX = "zenml"
ZENML_VAULT_SECRET_METADATA_KEY = "zenml_secret_metadata"
ZENML_VAULT_SECRET_CREATED_KEY = "zenml_secret_created"
ZENML_VAULT_SECRET_UPDATED_KEY = "zenml_secret_updated"


class HashiCorpVaultSecretsStoreBackend(BaseSecretsStoreBackend):
    """HashiCorp Vault ZenML secrets store backend."""

    def _convert_vault_secret(
        self,
        vault_secret: Dict[str, Any],
    ) -> ZenMLSecretMetadata:
        """Create a ZenML secret model from data stored in an HashiCorp Vault secret.

        If the HashiCorp Vault secret cannot be converted, the method acts as if
        the secret does not exist and raises a KeyError.

        Args:
            vault_secret: The HashiCorp Vault secret in JSON form.

        Returns:
            The ZenML secret.

        Raises:
            KeyError: if the HashiCorp Vault secret cannot be converted.
        """
        try:
            metadata = vault_secret[ZENML_VAULT_SECRET_METADATA_KEY]
            created = datetime.fromisoformat(
                vault_secret[ZENML_VAULT_SECRET_CREATED_KEY],
            )
            updated = datetime.fromisoformat(
                vault_secret[ZENML_VAULT_SECRET_UPDATED_KEY],
            )
        except (KeyError, ValueError) as e:
            raise KeyError(
                f"Secret could not be retrieved: missing required metadata: {e}"
            )

        return self._create_secret_from_metadata(
            metadata=metadata,
            created=created,
            updated=updated,
        )

    def list_secrets(self) -> List[ZenMLSecretMetadata]:
        """List all secrets.

        Note that returned secrets do not include any secret values. To fetch
        the secret values, use `get_secret`.

        Returns:
            A list of all secrets.

        Raises:
            RuntimeError: If the HashiCorp Vault API returns an unexpected
                error.
        """
        from hvac import Client
        from hvac.exceptions import (
            InvalidPath,
            VaultError,
        )

        assert isinstance(self.client, Client)

        results: List[ZenMLSecretMetadata] = []

        try:
            # List all ZenML secrets in the Vault
            all_secrets = (
                self.client.secrets.kv.v2.list_secrets(
                    path=HVAC_ZENML_SECRET_NAME_PREFIX
                )
                .get("data", {})
                .get("keys", [])
            )
        except InvalidPath:
            # no secrets created yet
            pass
        except VaultError as e:
            raise RuntimeError(f"Error listing HashiCorp Vault secrets: {e}")
        else:
            # Convert the Vault secrets to ZenML secrets
            for secret_uuid in all_secrets:
                vault_secret_id = (
                    f"{HVAC_ZENML_SECRET_NAME_PREFIX}/{secret_uuid}"
                )
                try:
                    vault_secret = (
                        self.client.secrets.kv.v2.read_secret(
                            path=vault_secret_id
                        )
                        .get("data", {})
                        .get("data", {})
                    )
                except (InvalidPath, VaultError) as e:
                    logger.warning(
                        f"Error fetching secret with ID {vault_secret_id}: {e}",
                    )
                    continue

                try:
                    secret_model = self._convert_vault_secret(
                        vault_secret,
                    )
                except KeyError as e:
                    # The _convert_vault_secret method raises a KeyError
                    # if the secret is not valid. Here we
                    # pretend that the secret does not exist.
                    logger.warning(
                        f"Error fetching secret with ID {vault_secret_id}: {e}",
                    )
                    continue

                results.append(secret_model)

        return results


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision.

    Raises:
        NotImplementedError: If the secrets store type is not supported.
    """
    from zenml.config.global_config import GlobalConfiguration
    from zenml.enums import SecretsStoreType
    from zenml.zen_stores.sql_zen_store import SqlZenStore

    logger.info("Migrating secrets from the external secrets store to the db.")

    store_cfg = GlobalConfiguration().store_configuration

    # We create a new SqlZenStore instance with the same configuration as the
    # one used to run migrations, but we skip the default registrations and
    # migrations. This is because we can't use the same store instance that is
    # used to run migrations because it would create an infinite loop.
    store = SqlZenStore(
        config=store_cfg,
        skip_default_registrations=True,
        skip_migrations=True,
    )

    secrets_store = store.secrets_store

    if secrets_store.TYPE == SecretsStoreType.SQL:
        # If the secrets store is already a SQL secrets store, we don't need
        # to transfer secrets from the external secrets store to the db.
        logger.info(
            "Skipping migration of secrets from the external secrets store "
            "to the db because the db is already configured as a SQL secrets "
            "store."
        )
        return

    logger.debug(
        f"Transferring secrets from the {secrets_store.type} secrets store "
        f"to the db: {secrets_store.config}"
    )

    # Transfer secrets from the external secrets store to the db

    conn = op.get_bind()
    meta = sa.MetaData()
    meta.reflect(
        only=("secret", "user", "workspace", "identity"), bind=op.get_bind()
    )
    secrets = sa.Table("secret", meta)
    users = sa.Table("user", meta)
    workspaces = sa.Table("workspace", meta)
    identity = sa.Table("identity", meta)

    # Extract the ZenML deployment ID from the identity table
    server_id = UUID(conn.execute(sa.select(identity.c.id)).scalar_one())

    # Initialize the secrets store backend based on the backend client extracted
    # from the secrets store.
    backend: BaseSecretsStoreBackend
    if secrets_store.TYPE == SecretsStoreType.AWS:
        from zenml.zen_stores.secrets_stores.aws_secrets_store import (
            AWSSecretsStore,
        )

        assert isinstance(secrets_store, AWSSecretsStore)
        backend = AWSSecretsStoreBackend(
            server_id=server_id,
            client=secrets_store.client,
        )
    elif secrets_store.TYPE == SecretsStoreType.GCP:
        from zenml.zen_stores.secrets_stores.gcp_secrets_store import (
            GCPSecretsStore,
        )

        assert isinstance(secrets_store, GCPSecretsStore)
        backend = GCPSecretsStoreBackend(
            server_id=server_id,
            client=secrets_store.client,
            project_id=secrets_store.config.project_id,
        )
    elif secrets_store.TYPE == SecretsStoreType.AZURE:
        from zenml.zen_stores.secrets_stores.azure_secrets_store import (
            AzureSecretsStore,
        )

        assert isinstance(secrets_store, AzureSecretsStore)
        backend = AzureSecretsStoreBackend(
            server_id=server_id,
            client=secrets_store.client,
        )
    elif secrets_store.TYPE == SecretsStoreType.HASHICORP:
        from zenml.zen_stores.secrets_stores.hashicorp_secrets_store import (
            HashiCorpVaultSecretsStore,
        )

        assert isinstance(secrets_store, HashiCorpVaultSecretsStore)
        backend = HashiCorpVaultSecretsStoreBackend(
            server_id=server_id,
            client=secrets_store.client,
        )
    else:
        raise NotImplementedError(
            f"Secrets store type {secrets_store.TYPE} not supported"
        )

    # List all secrets in the backend
    external_secrets = backend.list_secrets()
    if len(external_secrets) == 0:
        logger.debug("No secrets found in the external secrets store.")

    for secret in external_secrets:
        logger.info(
            f"Transferring secret '{secret.name}' from the "
            f"{secrets_store.type} secrets store to the db."
        )
        # Check if a user with the given ID exists in the db
        user = conn.execute(
            users.select().where(
                users.c.id == str(secret.user_id).replace("-", "")
            )
        ).first()
        if user is None:
            logger.warning(
                f"User with ID {secret.user_id} not found in the db. "
                f"Skipping secret '{secret.name}'."
            )
            continue
        # Check if a workspace with the given ID exists in the db
        workspace = conn.execute(
            workspaces.select().where(
                workspaces.c.id == str(secret.workspace_id).replace("-", "")
            )
        ).first()
        if workspace is None:
            logger.warning(
                f"Workspace with ID {secret.workspace_id} not found in the db. "
                f"Skipping secret '{secret.name}'."
            )
            continue

        # Check if there is already a secret with the same ID in the db
        # and delete it if there is. This can happen for example if the user has
        # previously migrated the secrets from the db to an external secrets
        # store and forgot to delete the secrets from the db.
        conn.execute(
            secrets.delete().where(
                secrets.c.id == str(secret.id).replace("-", "")
            )
        )
        conn.execute(
            secrets.insert().values(
                id=str(secret.id).replace("-", ""),
                created=secret.created,
                updated=secret.updated,
                name=secret.name,
                scope=secret.scope,
                values=None,
                workspace_id=str(secret.workspace_id).replace("-", ""),
                user_id=str(secret.user_id).replace("-", ""),
            )
        )


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass



# Copyright (c) ZenML GmbH 2026. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Registry module for type/class trigger resolution."""

from zenml.enums import TriggerType
from zenml.models.v2.core.triggers import (
    ScheduleTriggerResponse,
    ScheduleTriggerResponseBody,
)

# Type to return class mappings - should be extended with new type classes

TYPE_TO_RESPONSE_BODY_MAPPING = {
    TriggerType.SCHEDULE.value: ScheduleTriggerResponseBody,
}

TYPE_TO_RESPONSE_MAPPING = {
    TriggerType.SCHEDULE.value: ScheduleTriggerResponse,
}

# Union type objects - should be extended with the future type classes (e.g. Webhook)



#  Copyright (c) ZenML GmbH 2025. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""SQLModel implementation of idempotent API transaction tables."""

from datetime import datetime, timedelta
from typing import Any, Optional
from uuid import UUID

from sqlalchemy import TEXT, Column, LargeBinary
from sqlalchemy.dialects.mysql import MEDIUMBLOB
from sqlmodel import Field, SQLModel

from zenml.models import (
    ApiTransactionRequest,
    ApiTransactionResponse,
    ApiTransactionResponseBody,
    ApiTransactionUpdate,
)
from zenml.utils.time_utils import utc_now
from zenml.zen_stores.schemas.base_schemas import BaseSchema
from zenml.zen_stores.schemas.schema_utils import (
    build_foreign_key_field,
    build_index,
)
from zenml.zen_stores.schemas.user_schemas import UserSchema


class ApiTransactionSchema(BaseSchema, table=True):
    """SQL Model for API transactions.

    The result payload is stored in a separate table to keep this table
    small and fast for cleanup operations. Deleting rows with large blobs
    is expensive because the entire row must be copied to the undo log.
    """

    __tablename__ = "api_transaction"
    __table_args__ = (
        build_index(
            table_name=__tablename__,
            column_names=[
                "completed",
                "expired",
            ],
        ),
    )
    method: str
    url: str = Field(sa_column=Column(TEXT, nullable=False))
    completed: bool = Field(default=False)
    expired: Optional[datetime] = Field(default=None, nullable=True)

    user_id: UUID = build_foreign_key_field(
        source=__tablename__,
        target=UserSchema.__tablename__,
        source_column="user_id",
        target_column="id",
        ondelete="CASCADE",
        nullable=False,
    )

    @classmethod
    def from_request(
        cls, request: ApiTransactionRequest
    ) -> "ApiTransactionSchema":
        """Create a new API transaction from a request.

        Args:
            request: The API transaction request.

        Returns:
            The API transaction schema.
        """
        assert request.user is not None, "User must be set."
        return cls(
            id=request.transaction_id,
            user_id=request.user,
            method=request.method,
            url=request.url,
            completed=False,
        )

    def to_model(
        self,
        include_metadata: bool = False,
        include_resources: bool = False,
        **kwargs: Any,
    ) -> ApiTransactionResponse:
        """Convert the SQL model to a ZenML model.

        Args:
            include_metadata: Whether to include metadata in the response.
            include_resources: Whether to include resources in the response.
            **kwargs: Additional keyword arguments.

        Returns:
            The API transaction response.
        """
        response = ApiTransactionResponse(
            id=self.id,
            body=ApiTransactionResponseBody(
                method=self.method,
                url=self.url,
                created=self.created,
                updated=self.updated,
                user_id=self.user_id,
                completed=self.completed,
            ),
        )
        return response

    def update(self, update: ApiTransactionUpdate) -> "ApiTransactionSchema":
        """Update the API transaction.

        Args:
            update: The API transaction update.

        Returns:
            The API transaction schema.
        """
        self.updated = utc_now()
        self.expired = self.updated + timedelta(seconds=update.cache_time)
        return self


class ApiTransactionResultSchema(SQLModel, table=True):
    """SQL Model for API transaction results."""

    __tablename__ = "api_transaction_result"

    id: UUID = build_foreign_key_field(
        source=__tablename__,
        target=ApiTransactionSchema.__tablename__,
        source_column="id",
        target_column="id",
        ondelete="CASCADE",
        nullable=False,
        primary_key=True,
    )

    result: bytes = Field(
        sa_column=Column(
            LargeBinary().with_variant(MEDIUMBLOB, "mysql"),
            nullable=False,
        ),
    )
